from fjformer.pallas_operations.tpu.ring_attention.ring_attention import (
	ring_attention_standard as ring_attention_standard,
	ring_flash_attention_tpu as ring_flash_attention_tpu,
	ring_attention as ring_attention,
)



# Copied from https://github.com/LargeWorldModel/LWM/blob/main/lwm/ring_attention.py

"""This module contains ring attention forward and backward pass, supporting both blockwise computation and
TPU-compatible fused attention. It features blockwise computation for feedforward networks to reduce memory cost.
For more details, refer to 'RingAttention' at https://arxiv.org/abs/2305.19370 and 'Blockwise Parallel Transformers'
at https://arxiv.org/abs/2310.01889.
"""

import dataclasses
import functools
from functools import partial
from typing import Any, NamedTuple, Optional

import jax
import jax.lax as lax
import jax.numpy as jnp
from einops import rearrange
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu


def _ring_attention_fwd(
	q, k, v, attn_bias, segment_ids, axis_name, float32_logits, blockwise_kwargs
):
	if float32_logits:
		q, k = q.astype(jnp.float32), k.astype(jnp.float32)
	batch, q_len, num_heads, dim_per_head = q.shape
	batch, kv_len, num_heads, dim_per_head = k.shape
	numerator = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(q.dtype)
	denominator = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
	axis_size = lax.psum(1, axis_name)
	q_block_size, kv_block_size = (
		q_len,
		kv_len,
	)  # assumes this function is pre-sharded inside shard_map
	query_chunk_size = blockwise_kwargs["query_chunk_size"]
	key_chunk_size = blockwise_kwargs["key_chunk_size"]

	def scan_kv_block(carry, idx):
		prev_max_score, numerator, denominator, k, v = carry
		q_block_idx = lax.axis_index(axis_name)
		k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
		q_chunk_idx_start = q_block_idx * (q_block_size // query_chunk_size)
		k_chunk_idx_start = k_block_idx * (kv_block_size // key_chunk_size)
		numerator, denominator, max_score = _blockwise_attention_fwd(
			q,
			k,
			v,
			(numerator, denominator, prev_max_score),
			q_chunk_idx_start,
			k_chunk_idx_start,
			bias=attn_bias,
			segment_ids=segment_ids,
			**blockwise_kwargs,
		)
		k, v = map(
			lambda x: lax.ppermute(
				x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]
			),
			(k, v),
		)
		return (max_score, numerator, denominator, k, v), None

	prev_max_score = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)
	(max_score, numerator, denominator, _, _), _ = lax.scan(
		scan_kv_block,
		init=(prev_max_score, numerator, denominator, k, v),
		xs=jnp.arange(0, axis_size),
	)
	output = numerator / rearrange(denominator, "b h q -> b q h")[..., None]
	return output.astype(v.dtype), (
		output,
		q,
		k,
		v,
		attn_bias,
		segment_ids,
		denominator,
		max_score,
	)


def _ring_attention_bwd(axis_name, float32_logits, blockwise_kwargs, res, g):
	del float32_logits
	output, q, k, v, attn_bias, segment_ids, denominator, max_score = res
	batch, q_len, num_heads, dim_per_head = q.shape
	batch, kv_len, num_heads, dim_per_head = k.shape
	axis_size = lax.psum(1, axis_name)
	dq = jnp.zeros_like(q, dtype=q.dtype)
	dk = jnp.zeros_like(k, dtype=k.dtype)
	dv = jnp.zeros_like(v, dtype=k.dtype)
	query_chunk_size = blockwise_kwargs["query_chunk_size"]
	key_chunk_size = blockwise_kwargs["key_chunk_size"]
	q_block_size, kv_block_size = (
		q_len,
		kv_len,
	)  # assumes this function is pre-sharded inside shard_map

	def scan_kv_block(carry, idx):
		dq, dk, dv, k, v = carry
		q_block_idx = lax.axis_index(axis_name)
		k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
		q_chunk_idx_start = q_block_idx * (q_block_size // query_chunk_size)
		k_chunk_idx_start = k_block_idx * (kv_block_size // key_chunk_size)
		dq, dk, dv = _blockwise_attention_bwd(
			q,
			k,
			v,
			g,
			(dq, dk, dv, output, denominator, max_score),
			q_chunk_idx_start,
			k_chunk_idx_start,
			bias=attn_bias,
			segment_ids=segment_ids,
			**blockwise_kwargs,
		)
		k, v, dk, dv = map(
			lambda x: lax.ppermute(
				x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]
			),
			(k, v, dk, dv),
		)
		return (dq, dk, dv, k, v), None

	(dq, dk, dv, k, v), _ = lax.scan(
		scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size)
	)
	dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(k.dtype)
	return dq, dk, dv, None, None


@partial(jax.custom_vjp, nondiff_argnums=[5, 6, 7])
def ring_attention(
	q, k, v, attn_bias, segment_ids, axis_name, float32_logits, blockwise_kwargs
):
	y, _ = _ring_attention_fwd(
		q, k, v, attn_bias, segment_ids, axis_name, float32_logits, blockwise_kwargs
	)
	return y


ring_attention.defvjp(_ring_attention_fwd, _ring_attention_bwd)


def _ring_attention_standard_fwd(q, k, v, attn_mask, axis_name, float32_logits):
	if float32_logits:
		q, k = q.astype(jnp.float32), k.astype(jnp.float32)
	batch, q_len, num_heads, _ = q.shape
	batch, kv_len, num_heads, dim_per_head = k.shape
	numerator = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(q.dtype)
	denominator = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
	axis_size = lax.psum(1, axis_name)
	scale = jnp.sqrt(q.shape[-1])

	def scan_kv_block(carry, idx):
		prev_max_score, numerator, denominator, k, v = carry
		mask = lax.dynamic_slice_in_dim(
			attn_mask,
			(lax.axis_index(axis_name) - idx) % axis_size * kv_len,
			kv_len,
			axis=-1,
		)
		attn_weights = jnp.einsum("bqhd,bkhd->bhqk", q, k) / scale
		attn_weights = jnp.where(mask, attn_weights, jnp.finfo(attn_weights.dtype).min)
		max_score = jnp.maximum(prev_max_score, jnp.max(attn_weights, axis=-1))
		exp_weights = jnp.exp(attn_weights - max_score[..., None])
		correction = rearrange(jnp.exp(prev_max_score - max_score), "b h q -> b q h")[
			..., None
		]
		numerator = numerator * correction + jnp.einsum("bhqk,bkhd->bqhd", exp_weights, v)
		denominator = denominator * jnp.exp(prev_max_score - max_score) + jnp.sum(
			exp_weights, axis=-1
		)
		k, v = map(
			lambda x: lax.ppermute(
				x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]
			),
			(k, v),
		)
		return (max_score, numerator, denominator, k, v), None

	prev_max_score = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)
	(max_score, numerator, denominator, _, _), _ = lax.scan(
		scan_kv_block,
		init=(prev_max_score, numerator, denominator, k, v),
		xs=jnp.arange(0, axis_size),
	)
	output = numerator / rearrange(denominator, "b h q -> b q h")[..., None]
	return output.astype(v.dtype), (
		output,
		q,
		k,
		v,
		attn_mask,
		numerator,
		denominator,
		max_score,
	)


def _ring_attention_standard_bwd(axis_name, float32_logits, res, g):
	del float32_logits
	axis_size = lax.psum(1, axis_name)
	output, q, k, v, attn_mask, numerator, denominator, max_score = res
	dq = jnp.zeros_like(q, dtype=jnp.float32)
	dk = jnp.zeros_like(k, dtype=jnp.float32)
	dv = jnp.zeros_like(v, dtype=jnp.float32)
	batch, kv_len, num_heads, dim_per_head = k.shape
	scale = jnp.sqrt(q.shape[-1])

	def scan_kv_block(carry, idx):
		dq, dk, dv, k, v = carry
		mask = lax.dynamic_slice_in_dim(
			attn_mask,
			(lax.axis_index(axis_name) - idx) % axis_size * kv_len,
			kv_len,
			axis=-1,
		)
		attn_weights = jnp.einsum("bqhd,bkhd->bhqk", q, k) / scale
		attn_weights = jnp.where(mask, attn_weights, jnp.finfo(attn_weights.dtype).min)
		exp_weights = jnp.exp(attn_weights - max_score[..., None]) / denominator[..., None]
		ds = jnp.einsum("bqhd,bkhd->bhqk", g, v)
		dl = (ds - jnp.einsum("bqhd,bqhd->bhq", g, output)[..., None]) * exp_weights
		dq = dq + jnp.einsum("bhqk,bkhd->bqhd", dl, k) / scale
		dk = dk + jnp.einsum("bqhd,bhqk->bkhd", q, dl) / scale
		dv = dv + jnp.einsum("bhqk,bqhd->bkhd", exp_weights, g)
		k, v, dk, dv = map(
			lambda x: lax.ppermute(
				x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]
			),
			(k, v, dk, dv),
		)
		return (dq, dk, dv, k, v), None

	(dq, dk, dv, k, v), _ = lax.scan(
		scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size)
	)
	dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
	return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5])
def ring_attention_standard(q, k, v, attn_mask, axis_name, float32_logits=True):
	y, _ = _ring_attention_standard_fwd(q, k, v, attn_mask, axis_name, float32_logits)
	return y


ring_attention_standard.defvjp(
	_ring_attention_standard_fwd, _ring_attention_standard_bwd
)


def _blockwise_attention_fwd(
	q,
	k,
	v,
	carry,
	q_chunk_idx_start,
	k_chunk_idx_start,
	bias,
	segment_ids,
	causal,
	query_chunk_size,
	key_chunk_size,
	deterministic,
	dropout_rng,
	attn_pdrop,
	dtype,
	policy,
	precision,
	prevent_cse,
):
	batch, q_len, num_heads, dim_per_head = q.shape
	batch, kv_len, num_heads, dim_per_head = k.shape
	batch, kv_len, num_heads, dim_per_head = v.shape
	num_q = q_len // query_chunk_size
	num_kv = kv_len // key_chunk_size
	q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
	k = k.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
	v = v.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
	q, k, v = map(lambda x: jnp.moveaxis(x, 1, 0), (q, k, v))

	numerator, denominator, max_score = carry
	numerator = numerator.reshape(
		(batch, num_q, query_chunk_size, num_heads, dim_per_head)
	)
	numerator = jnp.moveaxis(numerator, 1, 0)
	denominator = denominator.reshape((batch, num_heads, num_q, query_chunk_size))
	max_score = max_score.reshape((batch, num_heads, num_q, query_chunk_size))
	denominator, max_score = map(
		lambda x: rearrange(x, "b h n c -> n b h c"), (denominator, max_score)
	)

	scale = jnp.sqrt(q.shape[-1])
	if not deterministic and attn_pdrop > 0.0:
		attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
		attn_dropout = jax.random.bernoulli(
			attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len)
		)
	else:
		attn_dropout = None
	_chunk_bias_fn = partial(
		_chunk_attention_bias,
		query_chunk_size,
		key_chunk_size,
		bias,
		segment_ids,
		deterministic,
		attn_dropout,
		attn_pdrop,
		causal,
		dtype,
	)

	def scan_attention(_, scan):
		q_chunk, numerator_chunk, denominator_chunk, max_score_chunk, q_chunk_idx = scan

		@partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
		def scan_kv_block(carry, scan):
			k_chunk, value_chunk, k_chunk_idx = scan
			numerator_chunk, denominator_chunk, prev_max_score_chunk = carry
			attn_weights = (
				jnp.einsum("bqhd,bkhd->bhqk", q_chunk, k_chunk, precision=precision) / scale
			)
			bias_chunk = _chunk_bias_fn(
				q_chunk_idx_start + q_chunk_idx, k_chunk_idx_start + k_chunk_idx
			)
			attn_weights = attn_weights + bias_chunk

			max_score_chunk = jnp.maximum(
				prev_max_score_chunk, jnp.max(attn_weights, axis=-1)
			)
			max_score_chunk = lax.stop_gradient(max_score_chunk)
			exp_weights = jnp.exp(attn_weights - max_score_chunk[..., None])
			exp_values = jnp.einsum(
				"bhqk,bkhd->bqhd", exp_weights, value_chunk, precision=precision
			)
			correction = rearrange(
				jnp.exp(prev_max_score_chunk - max_score_chunk), "b h q -> b q h"
			)[..., None]
			numerator_chunk = numerator_chunk * correction + exp_values
			denominator_chunk = denominator_chunk * jnp.exp(
				prev_max_score_chunk - max_score_chunk
			) + exp_weights.sum(axis=-1)
			return (numerator_chunk, denominator_chunk, max_score_chunk), None

		def skip_upper_half(carry, args):
			key_chunk, value_chunk, k_chunk_idx = args
			skip_block = jnp.array(False)
			if causal:
				skip_block = q_chunk_idx_start + q_chunk_idx < k_chunk_idx_start + k_chunk_idx
			return jax.lax.cond(
				skip_block,
				lambda carry, args: (carry, None),
				scan_kv_block,
				carry,
				args,
			)

		(numerator_chunk, denominator_chunk, max_score_chunk), _ = lax.scan(
			skip_upper_half,
			init=(numerator_chunk, denominator_chunk, max_score_chunk),
			xs=(k, v, jnp.arange(0, num_kv)),
		)
		output_chunk = numerator_chunk / rearrange(denominator_chunk, "b h q -> b q h")[
			..., None
		].astype(dtype)
		return (), (output_chunk, numerator_chunk, denominator_chunk, max_score_chunk)

	_, (_, numerator, denominator, max_score) = lax.scan(
		scan_attention,
		init=(),
		xs=(q, numerator, denominator, max_score, jnp.arange(0, num_q)),
	)

	numerator = jnp.moveaxis(numerator, 1, 0)
	numerator = numerator.reshape((batch, q_len, num_heads, dim_per_head))
	denominator, max_score = map(
		lambda x: rearrange(x, "n b h c -> b h n c"), (denominator, max_score)
	)
	denominator = denominator.reshape((batch, num_heads, q_len))
	max_score = max_score.reshape((batch, num_heads, q_len))

	return numerator, denominator, max_score


def _blockwise_attention_bwd(
	q,
	k,
	v,
	g,
	carry,
	q_chunk_idx_start,
	k_chunk_idx_start,
	bias,
	segment_ids,
	causal,
	query_chunk_size,
	key_chunk_size,
	deterministic,
	dropout_rng,
	attn_pdrop,
	dtype,
	policy,
	precision,
	prevent_cse,
):
	batch, q_len, num_heads, dim_per_head = q.shape
	batch, kv_len, num_heads, dim_per_head = k.shape
	batch, kv_len, num_heads, dim_per_head = v.shape
	num_q = q_len // query_chunk_size
	num_kv = kv_len // key_chunk_size
	dq, dk, dv, output, denominator, max_score = carry

	g = g.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
	dq = dq.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
	dk = dk.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
	dv = dv.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
	output = output.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
	g, dq, dk, dv, output = map(lambda x: jnp.moveaxis(x, 1, 0), (g, dq, dk, dv, output))

	denominator = denominator.reshape((batch, num_heads, num_q, query_chunk_size))
	max_score = max_score.reshape((batch, num_heads, num_q, query_chunk_size))
	denominator, max_score = map(
		lambda x: rearrange(x, "b h n c -> n b h c"), (denominator, max_score)
	)

	q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
	k = k.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
	v = v.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
	q, k, v = map(lambda x: jnp.moveaxis(x, 1, 0), (q, k, v))

	scale = jnp.sqrt(q.shape[-1])
	if not deterministic and attn_pdrop > 0.0:
		attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
		attn_dropout = jax.random.bernoulli(
			attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len)
		)
	else:
		attn_dropout = None
	_chunk_bias_fn = partial(
		_chunk_attention_bias,
		query_chunk_size,
		key_chunk_size,
		bias,
		segment_ids,
		deterministic,
		attn_dropout,
		attn_pdrop,
		causal,
		dtype,
	)

	def scan_attention(carry, scan):
		dk, dv = carry
		(
			q_chunk,
			dq_chunk,
			g_chunk,
			output_chunk,
			denominator_chunk,
			max_score_chunk,
			q_chunk_idx,
		) = scan
		dl_part = jnp.einsum("bqhd,bqhd->bhq", g_chunk, output_chunk)[..., None]

		@partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
		def scan_kv_block(carry, scan):
			k_chunk, value_chunk, k_chunk_idx = scan
			dq_chunk = carry
			attn_weights = (
				jnp.einsum("bqhd,bkhd->bhqk", q_chunk, k_chunk, precision=precision) / scale
			)
			bias_chunk = _chunk_bias_fn(
				q_chunk_idx_start + q_chunk_idx, k_chunk_idx_start + k_chunk_idx
			)
			attn_weights = attn_weights + bias_chunk
			exp_weights = (
				jnp.exp(attn_weights - max_score_chunk[..., None])
				/ denominator_chunk[..., None]
			)

			ds = jnp.einsum("bqhd,bkhd->bhqk", g_chunk, value_chunk)
			dl = (ds - dl_part) * exp_weights
			dq_chunk = dq_chunk + jnp.einsum("bhqk,bkhd->bqhd", dl, k_chunk) / scale
			dk_chunk = jnp.einsum("bqhd,bhqk->bkhd", q_chunk, dl) / scale
			dv_chunk = jnp.einsum("bhqk,bqhd->bkhd", exp_weights, g_chunk)
			return dq_chunk, (dk_chunk, dv_chunk)

		def skip_upper_half(carry, args):
			key_chunk, value_chunk, k_chunk_idx = args
			skip_block = jnp.array(False)
			if causal:
				skip_block = q_chunk_idx_start + q_chunk_idx < k_chunk_idx_start + k_chunk_idx
			return lax.cond(
				skip_block,
				lambda carry, args: (
					carry,
					(
						jnp.zeros(
							(batch, key_chunk_size, num_heads, dim_per_head),
							dtype=dk.dtype,
						),
						jnp.zeros(
							(batch, key_chunk_size, num_heads, dim_per_head),
							dtype=dk.dtype,
						),
					),
				),
				scan_kv_block,
				carry,
				args,
			)

		dq_chunk, (dk_part, dv_part) = lax.scan(
			skip_upper_half, init=dq_chunk, xs=(k, v, jnp.arange(0, num_kv))
		)
		return (dk + dk_part, dv + dv_part), dq_chunk

	(dk, dv), dq = lax.scan(
		scan_attention,
		init=(dk, dv),
		xs=(q, dq, g, output, denominator, max_score, jnp.arange(0, num_q)),
	)

	dq, dk, dv = map(lambda x: jnp.moveaxis(x, 1, 0), (dq, dk, dv))
	dq = dq.reshape((batch, q_len, num_heads, dim_per_head))
	dk = dk.reshape((batch, kv_len, num_heads, dim_per_head))
	dv = dv.reshape((batch, kv_len, num_heads, dim_per_head))

	return dq, dk, dv


def _chunk_attention_bias(
	query_chunk_size,
	key_chunk_size,
	bias,
	segment_ids,
	deterministic,
	attn_dropout,
	attn_pdrop,
	causal,
	dtype,
	query_chunk_idx,
	key_chunk_idx,
):
	query_offset = query_chunk_idx * query_chunk_size
	key_offset = key_chunk_idx * key_chunk_size
	chunk_bias = jnp.zeros((1, 1, 1, 1), dtype=dtype)
	if bias is not None:
		chunk_bias = lax.dynamic_slice(
			bias,
			start_indices=(0, 0, 0, key_offset),
			slice_sizes=(
				*bias.shape[:2],
				min(bias.shape[-2], query_chunk_size),
				min(bias.shape[-1], key_chunk_size),
			),
		)

	if segment_ids is not None:
		q_segment_ids = lax.dynamic_slice(
			segment_ids,
			start_indices=(0, query_offset),
			slice_sizes=(segment_ids.shape[0], query_chunk_size),
		)
		k_segment_ids = lax.dynamic_slice(
			segment_ids,
			start_indices=(0, key_offset),
			slice_sizes=(segment_ids.shape[0], key_chunk_size),
		)
		segment_ids_mask = q_segment_ids[:, :, None] != k_segment_ids[:, None, :]
		segment_ids_mask = segment_ids_mask[:, None]  # B1QK
		segment_ids_bias = segment_ids_mask * jnp.finfo(dtype).min
		chunk_bias += segment_ids_bias

	if causal:
		query_idx = lax.broadcasted_iota(
			dtype=jnp.int32, shape=(query_chunk_size, 1), dimension=0
		)
		key_idx = lax.broadcasted_iota(
			dtype=jnp.int32, shape=(1, key_chunk_size), dimension=1
		)
		offset = query_offset - key_offset
		query_idx += offset
		causal_mask_value = (query_idx < key_idx) * jnp.finfo(dtype).min
		chunk_bias += causal_mask_value.reshape(1, 1, *causal_mask_value.shape)

	if not deterministic and attn_pdrop > 0.0:
		attn_dropout_slice = lax.dynamic_slice(
			attn_dropout,
			start_indices=(0, 0, query_offset, key_offset),
			slice_sizes=(
				*attn_dropout.shape[:2],
				min(attn_dropout.shape[-2], query_chunk_size),
				min(attn_dropout.shape[-1], key_chunk_size),
			),
		)
		chunk_bias += attn_dropout_slice * jnp.finfo(dtype).min
	return chunk_bias.astype(dtype)


def _ring_flash_attention_fwd_tpu(
	q, k, v, attn_bias, segment_ids, axis_name, float32_logits, blockwise_kwargs
):
	if float32_logits:
		q, k = q.astype(jnp.float32), k.astype(jnp.float32)
	q, k, v = map(lambda x: rearrange(x, "b q h d -> b h q d"), [q, k, v])
	batch, num_heads, q_len, dim_per_head = q.shape
	batch, num_heads, kv_len, dim_per_head = k.shape
	attn_bias = attn_bias[:, 0, 0]  # (batch, k_len)

	o = jnp.zeros((batch, num_heads, q_len, dim_per_head)).astype(q.dtype)
	l = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
	m = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)

	axis_size = lax.psum(1, axis_name)
	q_block_size, kv_block_size = (
		q_len,
		kv_len,
	)  # assumes this function is pre-sharded inside shard_map
	query_chunk_size = blockwise_kwargs["query_chunk_size"]
	key_chunk_size = blockwise_kwargs["key_chunk_size"]
	if segment_ids is not None:
		q_segment_ids = lax.dynamic_slice_in_dim(
			segment_ids, lax.axis_index(axis_name) * q_len, q_len, axis=-1
		)

	block_sizes = BlockSizes(
		block_q=query_chunk_size,
		block_k_major=key_chunk_size,
		block_k=key_chunk_size,
		block_b=1,
		block_q_major_dkv=query_chunk_size,
		block_k_major_dkv=key_chunk_size,
		block_k_dkv=key_chunk_size,
		block_q_dkv=query_chunk_size,
		block_k_major_dq=key_chunk_size,
		block_k_dq=key_chunk_size,
		block_q_dq=query_chunk_size,
	)

	scale = q.shape[-1] ** -0.5

	def scan_kv_block(carry, idx):
		o, l, m, k, v = carry
		attn_bias_slice = lax.dynamic_slice_in_dim(
			attn_bias,
			(lax.axis_index(axis_name) - idx) % axis_size * kv_len,
			kv_len,
			axis=-1,
		)
		attn_bias_slice = None  # TODO
		if segment_ids is not None:
			kv_segment_ids = lax.dynamic_slice_in_dim(
				segment_ids,
				(lax.axis_index(axis_name) - idx) % axis_size * kv_len,
				kv_len,
				axis=-1,
			)
			segment_ids_slice = SegmentIds(q=q_segment_ids, kv=kv_segment_ids)
		else:
			segment_ids_slice = None
		q_block_idx = lax.axis_index(axis_name)
		k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
		q_chunk_idx_start = q_block_idx * (q_block_size // query_chunk_size)
		k_chunk_idx_start = k_block_idx * (kv_block_size // key_chunk_size)
		o, l, m = _flash_attention_fwd(
			q,
			k,
			v,
			carry=(o, l, m),
			q_chunk_idx_start=q_chunk_idx_start,
			k_chunk_idx_start=k_chunk_idx_start,
			ab=attn_bias_slice,
			segment_ids=segment_ids_slice,
			save_residuals=False,
			causal=blockwise_kwargs["causal"],
			sm_scale=scale,
			block_sizes=block_sizes,
			debug=False,
		)
		k, v = map(
			lambda x: lax.ppermute(
				x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]
			),
			(k, v),
		)
		return (o, l, m, k, v), None

	(o, l, m, _, _), _ = lax.scan(
		scan_kv_block, init=(o, l, m, k, v), xs=jnp.arange(0, axis_size)
	)
	output = rearrange(o.astype(v.dtype), "b h q d -> b q h d")
	return output, (o, q, k, v, attn_bias, segment_ids, l, m)


def _ring_flash_attention_bwd_tpu(axis_name, float32_logits, blockwise_kwargs, res, g):
	del float32_logits
	o, q, k, v, attn_bias, segment_ids, l, m = res
	batch, num_heads, kv_len, dim_per_head = k.shape
	axis_size = lax.psum(1, axis_name)
	dq = jnp.zeros_like(q, dtype=jnp.float32)
	dk = jnp.zeros_like(k, dtype=jnp.float32)
	dv = jnp.zeros_like(v, dtype=jnp.float32)
	query_chunk_size = blockwise_kwargs["query_chunk_size"]
	key_chunk_size = blockwise_kwargs["key_chunk_size"]
	q_block_size, kv_block_size = (
		q.shape[2],
		k.shape[2],
	)  # assumes this function is pre-sharded inside shard_map
	scale = q.shape[-1] ** -0.5

	if segment_ids is not None:
		q_segment_ids = lax.dynamic_slice_in_dim(
			segment_ids, lax.axis_index(axis_name) * q_block_size, q_block_size, axis=-1
		)

	g = rearrange(g, "b q h d -> b h q d")

	block_sizes = BlockSizes(
		block_q=query_chunk_size,
		block_k_major=key_chunk_size,
		block_k=key_chunk_size,
		block_b=1,
		block_q_major_dkv=query_chunk_size,
		block_k_major_dkv=key_chunk_size,
		block_k_dkv=key_chunk_size,
		block_q_dkv=query_chunk_size,
		block_k_major_dq=key_chunk_size,
		block_k_dq=key_chunk_size,
		block_q_dq=query_chunk_size,
	)

	def scan_kv_block(carry, idx):
		dq, dk, dv, k, v = carry
		attn_bias_slice = lax.dynamic_slice_in_dim(
			attn_bias,
			(lax.axis_index(axis_name) - idx) % axis_size * kv_len,
			kv_len,
			axis=-1,
		)
		attn_bias_slice = None  # TODO
		if segment_ids is not None:
			kv_segment_ids = lax.dynamic_slice_in_dim(
				segment_ids,
				(lax.axis_index(axis_name) - idx) % axis_size * kv_len,
				kv_len,
				axis=-1,
			)
			segment_ids_slice = SegmentIds(q=q_segment_ids, kv=kv_segment_ids)
		else:
			segment_ids_slice = None
		q_block_idx = lax.axis_index(axis_name)
		k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
		q_chunk_idx_start = q_block_idx * (q_block_size // query_chunk_size)
		k_chunk_idx_start = k_block_idx * (kv_block_size // key_chunk_size)
		(
			dq_i,
			dk_i,
			dv_i,
		) = _flash_attention_bwd(
			save_residuals=False,
			causal=blockwise_kwargs["causal"],
			sm_scale=scale,
			block_sizes=block_sizes,
			debug=False,
			q_chunk_idx_start=q_chunk_idx_start,
			k_chunk_idx_start=k_chunk_idx_start,
			residuals=(q, k, v, attn_bias_slice, segment_ids_slice, o, l, m),
			do=g,
		)
		dq += dq_i
		dk += dk_i
		dv += dv_i
		k, v, dk, dv = map(
			lambda x: lax.ppermute(
				x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]
			),
			(k, v, dk, dv),
		)
		return (dq, dk, dv, k, v), None

	(dq, dk, dv, k, v), _ = lax.scan(
		scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size)
	)
	dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
	dq, dk, dv = map(lambda x: rearrange(x, "b h q d -> b q h d"), (dq, dk, dv))
	return dq, dk, dv, None, None


@partial(jax.custom_vjp, nondiff_argnums=[5, 6, 7])
def ring_flash_attention_tpu(
	q, k, v, attn_bias, segment_ids, axis_name, float32_logits, blockwise_kwargs
):
	y, _ = _ring_flash_attention_fwd_tpu(
		q, k, v, attn_bias, segment_ids, axis_name, float32_logits, blockwise_kwargs
	)
	return y


ring_flash_attention_tpu.defvjp(
	_ring_flash_attention_fwd_tpu, _ring_flash_attention_bwd_tpu
)

# TPU-compatible fused attention functions for RingAttention
DEFAULT_MASK_VALUE = -0.7 * float(jnp.finfo(jnp.dtype("float32")).max)
NUM_LANES = 128
NUM_SUBLANES = 8


class SegmentIds(NamedTuple):
	"""SegmentIds for Q and KV sequences.

	SegmentIds are used to generate segment mask, which prevents attention between
	different segments in the input sequence. Each array is a list of ids
	(integers).
	Only the token with the same id can attend to each other.

	Attributes:
	  q: segment ids along the Q sequence.
	  kv: segment ids along the KV sequence.
	"""

	q: jax.Array  # [q_seq_len]
	kv: jax.Array  # [kv_seq_len]


@dataclasses.dataclass(frozen=True)
class BlockSizes:
	block_q: int
	block_k_major: int
	block_k: int
	block_b: int

	block_q_major_dkv: Optional[int] = None
	block_k_major_dkv: Optional[int] = None
	block_k_dkv: Optional[int] = None
	block_q_dkv: Optional[int] = None

	block_k_major_dq: Optional[int] = None
	block_k_dq: Optional[int] = None
	block_q_dq: Optional[int] = None

	def __post_init__(self):
		def verify_major_minor(prefix, suffix, major, minor):
			if minor > major:
				raise ValueError(
					f"{prefix}{suffix}={minor} should be smaller than"
					f" {prefix}_major{suffix}={major}"
				)
			if major % minor != 0:
				raise ValueError(
					f"{prefix}{suffix}={minor} should divide" f" {prefix}_major{suffix}={major}"
				)

		verify_major_minor("block_k", "", self.block_k_major, self.block_k)
		if self.block_q_major_dkv is not None and self.block_q_dkv is not None:
			verify_major_minor("block_q", "_dkv", self.block_q_major_dkv, self.block_q_dkv)
		if self.block_k_major_dkv is not None and self.block_k_dkv is not None:
			verify_major_minor("block_k", "_dkv", self.block_k_major_dkv, self.block_k_dkv)
		if self.block_k_major_dq is not None and self.block_k_dq is not None:
			verify_major_minor("block_k", "_dq", self.block_k_major_dq, self.block_k_dq)

	@property
	def has_backward_blocks(self) -> bool:
		backward_blocks = (
			self.block_q_major_dkv,
			self.block_k_major_dkv,
			self.block_q_dkv,
			self.block_k_dkv,
			self.block_k_major_dq,
			self.block_k_dq,
			self.block_q_dq,
		)
		return all(b is not None for b in backward_blocks)

	@classmethod
	def get_default(cls, batch_size, num_heads, q_seq_len, kv_len, d_model):
		del batch_size, num_heads, q_seq_len, kv_len, d_model  # Unused.
		return BlockSizes(
			block_q=128,
			block_k_major=128,
			block_k=128,
			block_b=1,
			block_q_major_dkv=128,
			block_k_major_dkv=128,
			block_k_dkv=128,
			block_q_dkv=128,
			block_k_major_dq=128,
			block_k_dq=128,
			block_q_dq=128,
		)


def _flash_attention(
	q,
	k,
	v,
	carry,
	q_chunk_idx_start,
	k_chunk_idx_start,
	ab,
	segment_ids,
	save_residuals,
	causal,
	sm_scale,
	block_sizes,
	debug,
):
	return _flash_attention_impl(
		q,
		k,
		v,
		carry,
		q_chunk_idx_start,
		k_chunk_idx_start,
		ab,
		segment_ids,
		save_residuals,
		causal,
		sm_scale,
		block_sizes.block_b,
		block_sizes.block_q,
		block_sizes.block_k_major,
		block_sizes.block_k,
		debug,
	)


def _flash_attention_fwd(
	q,
	k,
	v,
	carry,
	q_chunk_idx_start,
	k_chunk_idx_start,
	ab,
	segment_ids,
	save_residuals,
	causal,
	sm_scale,
	block_sizes,
	debug,
):
	if save_residuals:
		raise NotImplementedError("Higher-order AD not supported")
	o, l, m = _flash_attention(
		q,
		k,
		v,
		carry,
		q_chunk_idx_start,
		k_chunk_idx_start,
		ab,
		segment_ids,
		True,
		causal,
		sm_scale,
		block_sizes,
		debug,
	)
	return o, l, m


def _flash_attention_bwd(
	save_residuals: bool,
	causal: bool,
	sm_scale: float,
	block_sizes: BlockSizes,
	debug: bool,
	q_chunk_idx_start,
	k_chunk_idx_start,
	residuals,
	do,
):
	"""VJP rule for FlashAttention."""
	if save_residuals:
		raise NotImplementedError("Higher-order AD not supported")
	(q, k, v, ab, segment_ids, o, l, m) = residuals
	if not block_sizes.has_backward_blocks:
		raise ValueError(
			"Program is being differentiated, but not all backward blocks are" " specified"
		)

	di = jnp.sum(
		o.astype(jnp.float32) * do.astype(jnp.float32), axis=-1
	)  # [batch_size, num_heads, q_seq_len]

	dk, dv = _flash_attention_bwd_dkv(
		q_chunk_idx_start,
		k_chunk_idx_start,
		q,
		k,
		v,
		ab,
		segment_ids,
		l,
		m,
		do,
		di,
		block_q_major=block_sizes.block_q_major_dkv,
		block_k_major=block_sizes.block_k_major_dkv,
		block_k=block_sizes.block_k_dkv,
		block_q=block_sizes.block_q_dkv,
		sm_scale=sm_scale,
		causal=causal,
		mask_value=DEFAULT_MASK_VALUE,
		debug=debug,
	)

	dq, ds = _flash_attention_bwd_dq(
		q_chunk_idx_start,
		k_chunk_idx_start,
		q,
		k,
		v,
		ab,
		segment_ids,
		l,
		m,
		do,
		di,
		block_q_major=block_sizes.block_q_dq,
		block_k_major=block_sizes.block_k_major_dq,
		block_k=block_sizes.block_k_dq,
		sm_scale=sm_scale,
		causal=causal,
		mask_value=DEFAULT_MASK_VALUE,
		debug=debug,
	)
	return dq, dk, dv


MIN_BLOCK_SIZE = 128
TRANS_B_DIM_NUMBERS = (((1,), (1,)), ((), ()))


def below_or_on_diag(r, r_blk_size, c, c_blk_size):
	# A block is considered below or on diagonal as long as the bottom left
	# corner of the block is below or on diagonal.
	return ((r + 1) * r_blk_size - 1) > (c * c_blk_size)


def _flash_attention_kernel(
	q_idx_chunk_start, k_idx_chunk_start, q_tile_ref, *args, **kwargs
):
	block_b = q_tile_ref.shape[0]
	# If we're not going to tile the softmax, then we can avoid a bunch of VPU ops.
	if kwargs["block_k"] == kwargs["kv_seq_len"]:
		raise AssertionError("Crash") from None
		kernel = ...
	else:
		kernel = _flash_attention_kernel_single_batch
	for batch_idx in range(block_b):
		kernel(
			(batch_idx, 0),
			q_idx_chunk_start,
			k_idx_chunk_start,
			q_tile_ref,
			*args,
			**kwargs,
		)


def _flash_attention_kernel_single_batch(
	batch_idx: tuple[int, ...],
	q_chunk_idx_start_ref,
	k_chunk_idx_start_ref,
	q_tile_ref,
	k_tile_ref,
	v_tile_ref,
	acc_tile_ref,
	l_tile_ref,
	m_tile_ref,
	ab_tile_ref,
	q_segment_ids_tile_ref,
	kv_segment_ids_tile_ref,  # Input arrays
	o_tile_ref,  # Output arrays
	m_scratch_ref,
	l_scratch_ref,
	acc_scratch_ref,
	l_ref: Optional[Any] = None,
	m_ref: Optional[Any] = None,
	*,
	causal,
	sm_scale,
	block_k,
	kv_seq_len,
	mask_value,
	block_q,
):
	block_k_major = k_tile_ref.shape[2]
	block_q = q_tile_ref.shape[2]
	head_dim = q_tile_ref.shape[-1]

	kv_seq_idx = pl.program_id(3)

	@pl.when(kv_seq_idx == 0)
	def start_new_sequence():
		m_scratch_ref[batch_idx] = m_tile_ref[batch_idx]
		l_scratch_ref[batch_idx] = l_tile_ref[batch_idx]
		acc_scratch_ref[batch_idx] = acc_tile_ref[batch_idx]

	q_chunk_idx_start = q_chunk_idx_start_ref[0]
	k_chunk_idx_start = k_chunk_idx_start_ref[0]

	q_seq_idx = pl.program_id(2)
	if causal:
		should_run = below_or_on_diag(
			q_seq_idx + q_chunk_idx_start,
			block_q,
			kv_seq_idx + k_chunk_idx_start,
			block_k_major,
		)
	else:
		should_run = True

	@pl.when(should_run)
	def run():
		@functools.partial(
			lax.fori_loop, 0, block_k_major // block_k, init_val=None, unroll=True
		)
		def body(i, _):
			m_prev = m_scratch_ref[batch_idx]
			l_prev = l_scratch_ref[batch_idx]
			q = q_tile_ref[batch_idx]  # [block_q, head_dim]
			start_k = i * block_k
			k = pl.load(
				k_tile_ref, (*batch_idx, pl.dslice(start_k, block_k), slice(None))
			)  # [block_k, head_dim]

			s = jax.lax.dot_general(
				q, k, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
			)  # [block_q, block_k]

			# Add attention bias if needed.
			if ab_tile_ref is not None:
				ab = pl.load(
					ab_tile_ref,
					(batch_idx[0], pl.dslice(0, block_q), pl.dslice(start_k, block_k)),
				).astype(jnp.float32)
				s += ab

			if sm_scale != 1.0:
				s *= sm_scale

			mask = None
			if q_segment_ids_tile_ref is not None:
				repeats, rem = divmod(block_k, NUM_LANES)
				if rem:
					raise NotImplementedError(f"kv block size must be a multiple of {NUM_LANES}")
				q_segment_ids = pltpu.repeat(
					q_segment_ids_tile_ref[batch_idx[0]], repeats, axis=1
				)  # [block_q, block_k].
				kv_segment_ids = pl.load(
					kv_segment_ids_tile_ref,
					(batch_idx[0], pl.dslice(1), pl.dslice(start_k, block_k)),
				)  # [1, block_k].
				mask = jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)

			if causal:
				mask_shape = (block_q, block_k)
				row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
				row_ids += (q_seq_idx + q_chunk_idx_start) * block_q
				col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
				col_ids += (kv_seq_idx + k_chunk_idx_start) * block_k_major + start_k
				causal_mask = col_ids <= row_ids
				mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)

			s = s if mask is None else s + jnp.where(mask, 0.0, mask_value)

			m_curr = jnp.max(s, axis=1)[:, None]  # Row max, shape [block_q, 1].
			m_next = jnp.maximum(m_prev, m_curr)  # Shape [block_q, 128].

			block_k_repeats, rem = divmod(block_k, MIN_BLOCK_SIZE)
			if rem:
				raise NotImplementedError(
					f"{block_k=} should be a multiple of {MIN_BLOCK_SIZE}"
				)
			p = jnp.exp(s - pltpu.repeat(m_next, block_k_repeats, 1))

			alpha = jnp.exp(m_prev - m_next)  # Shape [block_q, 128].

			l_corr = alpha * l_prev

			l_next = jnp.sum(p, axis=1)[:, None] + l_corr  # Shape [block_q, 128]

			head_dim_repeats, rem = divmod(head_dim, MIN_BLOCK_SIZE)
			l_broadcast = lambda l: pltpu.repeat(l, head_dim_repeats, 1)
			if rem:
				if head_dim_repeats == 0:
					l_broadcast = lambda l: l[:, :head_dim]
				else:
					raise NotImplementedError(
						f"{head_dim=} should be a multiple of {MIN_BLOCK_SIZE} if larger"
					)
			l_scratch_ref[batch_idx] = l_next
			m_scratch_ref[batch_idx] = m_next

			l_next_inv_safe = jnp.where(l_next == 0.0, 1.0, 1.0 / l_next)
			acc_scratch_ref[batch_idx] *= l_broadcast(l_corr * l_next_inv_safe)
			v = pl.load(v_tile_ref, (*batch_idx, pl.dslice(start_k, block_k), slice(None)))
			o_curr = jax.lax.dot(p.astype(v.dtype), v, preferred_element_type=jnp.float32)
			acc_scratch_ref[batch_idx] += o_curr * l_broadcast(l_next_inv_safe)

	@pl.when(kv_seq_idx == (kv_seq_len // block_k_major) - 1)
	def store_output():
		o_tile_ref[batch_idx] = acc_scratch_ref[batch_idx].astype(o_tile_ref.dtype)
		if l_ref is not None:
			l_ref[batch_idx] = l_scratch_ref[batch_idx].astype(l_ref.dtype)
		if m_ref is not None:
			m_ref[batch_idx] = m_scratch_ref[batch_idx].astype(m_ref.dtype)


def _flash_attention_impl(
	q,
	k,
	v,
	carry,
	q_chunk_idx_start,
	k_chunk_idx_start,
	ab,
	segment_ids,
	save_residuals,
	causal,
	sm_scale,
	block_b,
	block_q,
	block_k_major,
	block_k,
	debug,
):
	assert block_k_major == block_k, (block_k_major, block_k)
	batch_size, num_heads, q_seq_len, head_dim = q.shape
	_, _, kv_seq_len, _ = k.shape
	acc, l_prev, m_prev = carry
	l_prev, m_prev = map(
		lambda x: jnp.broadcast_to(x[..., None], (*x.shape, MIN_BLOCK_SIZE)),
		(l_prev, m_prev),
	)
	q_chunk_idx_start, k_chunk_idx_start = (
		q_chunk_idx_start[None],
		k_chunk_idx_start[None],
	)
	_verify_block("block_q", "q_seq_len", block_q, q_seq_len, should_divide=False)
	_verify_block("block_k_major", "kv_seq_len", block_k_major, kv_seq_len)
	_verify_block("block_k", "kv_seq_len", block_k, kv_seq_len)
	_verify_block("block_b", "batch", block_b, batch_size, should_divide=False)

	grid = (
		pl.cdiv(batch_size, block_b),
		num_heads,
		pl.cdiv(q_seq_len, block_q),
		kv_seq_len // block_k_major,
	)

	def q_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, q_seq_index, 0)

	def kv_index_map(
		batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref
	):
		if causal:
			# If the kv block is skipped, prefetch the next valid kv block, i.e. the
			# 0th one to be used for the next block_q rows.
			next_kv_index = lax.select(
				below_or_on_diag(
					q_seq_index + q_idx_ref[0],
					block_q,
					kv_seq_index + k_idx_ref[0],
					block_k_major,
				),
				kv_seq_index,
				0,
			)
		else:
			next_kv_index = kv_seq_index
		return (batch_index, head_index, next_kv_index, 0)

	def ab_index_map(
		batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref
	):
		if causal:
			should_run = below_or_on_diag(
				q_seq_index + q_idx_ref[0],
				block_q,
				kv_seq_index + k_idx_ref[0],
				block_k_major,
			)
			next_kv_index = lax.select(should_run, kv_seq_index, 0)
		else:
			next_kv_index = kv_seq_index

		return (batch_index, 0, next_kv_index)

	def o_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, q_seq_index, 0)

	def lm_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, q_seq_index, 0)

	kernel = functools.partial(
		_flash_attention_kernel,
		causal=causal,
		mask_value=DEFAULT_MASK_VALUE,
		sm_scale=sm_scale,
		block_k=block_k,
		kv_seq_len=kv_seq_len,
		block_q=block_q,
	)
	out_shape = [jax.ShapeDtypeStruct(shape=q.shape, dtype=q.dtype)]
	out_specs = [pl.BlockSpec((block_b, 1, block_q, head_dim), o_index_map)]

	if block_k != kv_seq_len:
		scratch_shape = functools.partial(jax.ShapeDtypeStruct, dtype=jnp.float32)
		m_scratch = scratch_shape((block_b, 1, block_q, MIN_BLOCK_SIZE))
		l_scratch = scratch_shape((block_b, 1, block_q, MIN_BLOCK_SIZE))
		acc_scratch = scratch_shape((block_b, 1, block_q, head_dim))
		out_shape += [m_scratch, l_scratch, acc_scratch]
		out_specs += [
			pl.BlockSpec(m_scratch.shape, lambda *_: (0, 0, 0, 0)),
			pl.BlockSpec(l_scratch.shape, lambda *_: (0, 0, 0, 0)),
			pl.BlockSpec(acc_scratch.shape, lambda *_: (0, 0, 0, 0)),
		]
	else:
		assert False
		out_shape += [None, None, None]
		out_specs += [None, None, None]

	if save_residuals:
		out_specs = [
			*out_specs,
			pl.BlockSpec((block_b, 1, block_q, MIN_BLOCK_SIZE), lm_index_map),
			pl.BlockSpec((block_b, 1, block_q, MIN_BLOCK_SIZE), lm_index_map),
		]
		l = jax.ShapeDtypeStruct(
			(batch_size, num_heads, q_seq_len, MIN_BLOCK_SIZE), dtype=jnp.float32
		)
		m = jax.ShapeDtypeStruct(
			(batch_size, num_heads, q_seq_len, MIN_BLOCK_SIZE), dtype=jnp.float32
		)
		out_shape = (*out_shape, l, m)

	ab_block_spec = (
		pl.BlockSpec(
			(block_b, block_q, block_k_major),
			ab_index_map,
		)
		if ab is not None
		else None
	)

	if ab is not None:
		ab = ab[:, None].repeat(block_q, axis=1)

	q_segment_ids_spec = kv_segment_ids_spec = None
	q_segment_ids = kv_segment_ids = None
	if segment_ids is not None:

		def q_segment_ids_index_map(
			batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref
		):
			del head_index
			return (batch_index, q_seq_index, 0)

		def kv_segment_ids_index_map(
			batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref
		):
			del head_index
			if causal:
				next_kv_index = lax.select(
					below_or_on_diag(
						q_seq_index + q_idx_ref[0],
						block_q,
						kv_seq_index + k_idx_ref[0],
						block_k_major,
					),
					kv_seq_index,
					0,
				)
			else:
				next_kv_index = kv_seq_index
			return (batch_index, 0, next_kv_index)

		q_segment_ids_spec = pl.BlockSpec(
			(block_b, block_q, NUM_LANES),
			q_segment_ids_index_map,
		)
		kv_segment_ids_spec = pl.BlockSpec(
			(block_b, NUM_SUBLANES, block_k_major),
			kv_segment_ids_index_map,
		)

		q_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.q,
			(batch_size, q_seq_len, NUM_LANES),
			(
				0,
				1,
			),
		)
		kv_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.kv,
			(batch_size, NUM_SUBLANES, kv_seq_len),
			(
				0,
				2,
			),
		)

	in_specs = [
		pl.BlockSpec((block_b, 1, block_q, head_dim), q_index_map),
		pl.BlockSpec((block_b, 1, block_k_major, head_dim), kv_index_map),
		pl.BlockSpec((block_b, 1, block_k_major, head_dim), kv_index_map),
		pl.BlockSpec((block_b, 1, block_q, head_dim), q_index_map),
		pl.BlockSpec((block_b, 1, block_q, MIN_BLOCK_SIZE), lm_index_map),
		pl.BlockSpec((block_b, 1, block_q, MIN_BLOCK_SIZE), lm_index_map),
		ab_block_spec,
		q_segment_ids_spec,
		kv_segment_ids_spec,
	]

	o, *aux = pl.pallas_call(
		kernel,
		out_shape=out_shape,
		grid_spec=pltpu.PrefetchScalarGridSpec(
			num_scalar_prefetch=2, in_specs=in_specs, out_specs=out_specs, grid=grid
		),
		debug=debug,
		mosaic_params=dict(
			dimension_semantics=("parallel", "parallel", "parallel", "arbitrary")
		),
	)(
		q_chunk_idx_start,
		k_chunk_idx_start,
		q,
		k,
		v,
		acc,
		l_prev,
		m_prev,
		ab,
		q_segment_ids,
		kv_segment_ids,
	)
	if save_residuals:
		l, m = (v[..., 0] for v in aux[-2:])
		return (o, l, m)
	else:
		return o


def _flash_attention_dkv_kernel(
	q_chunk_idx_start_ref,
	k_chunk_idx_start_ref,
	q_tile_ref,
	k_tile_ref,
	v_tile_ref,
	ab_tile_ref,
	q_segment_ids_tile_ref,
	kv_segment_ids_tile_ref,
	l_tile_ref,
	m_tile_ref,
	do_tile_ref,
	di_tile_ref,
	dk_tile_ref,
	dv_tile_ref,
	dk_scratch_ref,
	dv_scratch_ref,
	*,
	sm_scale: float,
	causal: bool,
	mask_value: float,
	q_seq_len: int,
	block_q: int,
	block_k: int,
):
	_, _, block_q_major, _ = q_tile_ref.shape
	_, _, block_k_major, _ = k_tile_ref.shape

	q_seq_index = pl.program_id(axis=3)
	kv_seq_index = pl.program_id(axis=2)

	q_chunk_idx_start = q_chunk_idx_start_ref[0]
	k_chunk_idx_start = k_chunk_idx_start_ref[0]

	@pl.when(q_seq_index == 0)
	def start_new_sequence():
		dk_scratch_ref[:, :] = jnp.zeros(dk_scratch_ref.shape, dk_scratch_ref.dtype)
		dv_scratch_ref[:, :] = jnp.zeros(dv_scratch_ref.shape, dv_scratch_ref.dtype)

	def q_body(j, _):
		start_q = j * block_q

		def k_body(i, _):
			start_k = i * block_k
			k = pl.load(k_tile_ref, (0, 0, pl.ds(start_k, block_k), slice(None)))
			v = pl.load(v_tile_ref, (0, 0, pl.ds(start_k, block_k), slice(None)))
			q = pl.load(
				q_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
			)  # [block_q, head_dim]
			l = pl.load(
				l_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
			)  # [block_q, 128]
			m = pl.load(
				m_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
			)  # [block_q, 128]
			do = pl.load(
				do_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
			)  # [block_q, 128]
			di = pl.load(di_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))).astype(
				jnp.float32
			)  # [block_q, 128]

			capped_logits = lax.dot_general(
				q, k, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
			)  # [block_q_major, block_k]

			if ab_tile_ref is not None:
				ab = pl.load(
					ab_tile_ref,
					(
						0,
						pl.dslice(0, block_q),
						pl.dslice(i * block_k, block_k),
					),
				).astype(jnp.float32)
				capped_logits += ab

			if sm_scale != 1.0:
				capped_logits *= sm_scale

			mask = None
			if q_segment_ids_tile_ref is not None:
				repeats, rem = divmod(block_k, NUM_LANES)
				if rem:
					raise NotImplementedError()
				q_segment_ids = pl.load(
					q_segment_ids_tile_ref, (0, pl.ds(start_q, block_q), slice(None))
				)  # [block_q, NUM_LANES].
				q_segment_ids = pltpu.repeat(
					q_segment_ids, repeats, axis=1
				)  # [block_q, block_k].
				kv_segment_ids = pl.load(
					kv_segment_ids_tile_ref, (slice(None), 0, pl.ds(start_k, block_k))
				)  # [1, block_k].
				mask = jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)

			if causal:
				mask_shape = (block_q, block_k)
				row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
				row_ids += (q_seq_index + q_chunk_idx_start) * block_q_major + start_q
				col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
				col_ids += (kv_seq_index + k_chunk_idx_start) * block_k_major + start_k
				causal_mask = col_ids <= row_ids
				mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)

			capped_logits = (
				capped_logits
				if mask is None
				else capped_logits + jnp.where(mask, 0.0, mask_value)
			)

			p = jnp.exp(capped_logits - pltpu.repeat(m, block_k // MIN_BLOCK_SIZE, axis=1))
			p = p * pltpu.repeat(
				1 / l, block_k // MIN_BLOCK_SIZE, axis=1
			)  # [block_q_major, block_k_major]
			dv = lax.dot(p.T.astype(do.dtype), do, preferred_element_type=jnp.float32)
			pl.store(
				dv_scratch_ref,
				(pl.ds(start_k, block_k), slice(None)),
				pl.load(dv_scratch_ref, (pl.ds(start_k, block_k), slice(None)))
				+ dv.astype(dv_scratch_ref.dtype),
			)

			# di: [block_q, 128]
			# do: [block_q, head_dim]
			# v: [block_k_major, head_dim]
			dp = lax.dot_general(
				do, v, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
			)
			ds = (dp - pltpu.repeat(di, block_k // MIN_BLOCK_SIZE, axis=1)) * p

			if sm_scale != 1.0:
				ds = ds * sm_scale

			# ds: [block_q_major, block_k_major]
			# q: [block_q_major, head_dim]
			dk = lax.dot(ds.T.astype(do.dtype), q, preferred_element_type=jnp.float32)
			pl.store(
				dk_scratch_ref,
				(pl.ds(start_k, block_k), slice(None)),
				pl.load(dk_scratch_ref, (pl.ds(start_k, block_k), slice(None)))
				+ dk.astype(dk_scratch_ref.dtype),
			)

		lax.fori_loop(0, block_k_major // block_k, k_body, None, unroll=True)

	if causal:
		should_run = below_or_on_diag(
			q_seq_index + q_chunk_idx_start,
			block_q_major,
			kv_seq_index + k_chunk_idx_start,
			block_k_major,
		)
	else:
		should_run = True

	@pl.when(should_run)
	def run():
		lax.fori_loop(0, block_q_major // block_q, q_body, None, unroll=True)

	@pl.when(q_seq_index == q_seq_len // block_q_major - 1)
	def end_of_q_sequence():
		dv_tile_ref[0, 0, :, :] = dv_scratch_ref[...].astype(dv_tile_ref)
		dk_tile_ref[0, 0, :, :] = dk_scratch_ref[...].astype(dk_tile_ref)


def _flash_attention_bwd_dkv(
	q_chunk_idx_start,
	k_chunk_idx_start,
	q,
	k,
	v,
	ab,
	segment_ids,
	l,
	m,
	do,
	di,
	*,
	block_q_major: Optional[int],
	block_q: Optional[int],
	block_k_major: Optional[int],
	block_k: Optional[int],
	sm_scale: float,
	causal: bool = False,
	mask_value: float = DEFAULT_MASK_VALUE,
	debug: bool = False,
):
	batch_size, num_heads, q_seq_len, head_dim = q.shape
	_, _, kv_seq_len, _ = k.shape
	q_chunk_idx_start, k_chunk_idx_start = (
		q_chunk_idx_start[None],
		k_chunk_idx_start[None],
	)
	_verify_block("block_q_major_dkv", "q_seq_len", block_q_major, q_seq_len)
	_verify_block("block_q_dkv", "q_seq_len", block_q, q_seq_len)
	_verify_block("block_k_major_dkv", "kv_seq_len", block_k_major, kv_seq_len)
	_verify_block("block_k_dkv", "kv_seq_len", block_k, kv_seq_len)

	# Broadcast out scalar values
	m = jnp.broadcast_to(m[..., None], (*m.shape, MIN_BLOCK_SIZE))
	l = jnp.broadcast_to(l[..., None], (*l.shape, MIN_BLOCK_SIZE))
	# Preprocess contraction for bwd pass
	di = jnp.broadcast_to(di[..., None], (*di.shape, MIN_BLOCK_SIZE))

	# kv index needs to be before q index since q index is the contractng
	# dimension.
	grid = (
		batch_size,
		num_heads,
		kv_seq_len // block_k_major,
		q_seq_len // block_q_major,
	)

	def qo_index_map(
		batch_index, head_index, kv_seq_index, q_seq_index, q_idx_ref, k_idx_ref
	):
		if causal:
			# If the q block is skipped, stay at the 0th q block.
			next_q_index = lax.select(
				below_or_on_diag(
					q_seq_index + q_idx_ref[0],
					block_q_major,
					kv_seq_index + k_idx_ref[0],
					block_k_major,
				),
				q_seq_index,
				0,
			)
		else:
			next_q_index = q_seq_index

		return (batch_index, head_index, next_q_index, 0)

	qo_spec = pl.BlockSpec((1, 1, block_q_major, head_dim), qo_index_map)
	assert qo_spec.block_shape is not None
	assert q.ndim == len(qo_spec.block_shape)
	do_spec = qo_spec
	assert do.ndim == len(qo_spec.block_shape)

	def kv_index_map(batch_index, head_index, kv_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, kv_seq_index, 0)

	kv_spec = pl.BlockSpec((1, 1, block_k_major, head_dim), kv_index_map)
	assert kv_spec.block_shape is not None
	assert k.ndim == len(kv_spec.block_shape)
	assert v.ndim == len(kv_spec.block_shape)

	def lm_index_map(batch_index, head_index, _, q_seq_index, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, q_seq_index, 0)

	lm_spec = pl.BlockSpec((1, 1, block_q_major, MIN_BLOCK_SIZE), lm_index_map)
	assert lm_spec.block_shape is not None
	assert l.ndim == len(lm_spec.block_shape)
	assert m.ndim == len(lm_spec.block_shape)

	di_spec = pl.BlockSpec((1, 1, block_q_major, MIN_BLOCK_SIZE), qo_index_map)
	assert di_spec.block_shape is not None
	assert di.ndim == len(di_spec.block_shape)

	def ab_index_map(
		batch_index, head_index, kv_seq_index, q_seq_index, q_idx_ref, k_idx_ref
	):
		return (batch_index, 0, kv_seq_index)

	if ab is not None:
		ab = ab[:, None].repeat(block_q_major, axis=1)

	dab_spec = (
		pl.BlockSpec((1, block_q_major, block_k_major), ab_index_map)
		if ab is not None
		else None
	)

	q_segment_ids_spec = kv_segment_ids_spec = None
	q_segment_ids = kv_segment_ids = None
	if segment_ids is not None:

		def q_segment_ids_index_map(
			batch_index, head_index, kv_seq_index, q_seq_index, q_idx_ref, k_idx_ref
		):
			del head_index
			if causal:
				next_q_index = lax.select(
					below_or_on_diag(
						q_seq_index + q_idx_ref[0],
						block_q_major,
						kv_seq_index + k_idx_ref[0],
						block_k_major,
					),
					q_seq_index,
					0,
				)
			else:
				next_q_index = q_seq_index
			return (batch_index, next_q_index, 0)

		def kv_segment_ids_index_map(
			batch_index, head_index, kv_seq_index, _, q_idx_ref, k_idx_ref
		):
			del head_index
			return (batch_index, 0, kv_seq_index)

		q_segment_ids_spec = pl.BlockSpec(
			(1, block_q_major, NUM_LANES),
			q_segment_ids_index_map,
		)
		kv_segment_ids_spec = pl.BlockSpec(
			(1, NUM_SUBLANES, block_k_major),
			kv_segment_ids_index_map,
		)

		q_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.q,
			(batch_size, q_seq_len, NUM_LANES),
			(
				0,
				1,
			),
		)
		kv_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.kv,
			(batch_size, NUM_SUBLANES, kv_seq_len),
			(
				0,
				2,
			),
		)

	in_specs = [
		qo_spec,
		kv_spec,
		kv_spec,
		dab_spec,
		q_segment_ids_spec,
		kv_segment_ids_spec,
		lm_spec,
		lm_spec,
		do_spec,
		di_spec,
	]

	out_shapes = [
		jax.ShapeDtypeStruct((batch_size, num_heads, kv_seq_len, head_dim), k.dtype),
		jax.ShapeDtypeStruct((batch_size, num_heads, kv_seq_len, head_dim), v.dtype),
		jax.ShapeDtypeStruct((block_k_major, head_dim), jnp.float32),
		jax.ShapeDtypeStruct((block_k_major, head_dim), jnp.float32),
	]

	def dkv_index_map(batch_index, head_index, kv_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, kv_seq_index, 0)

	dkv_spec = pl.BlockSpec((1, 1, block_k_major, head_dim), dkv_index_map)
	out_specs = [
		dkv_spec,
		dkv_spec,
		pl.BlockSpec((block_k_major, head_dim), lambda *_: (0, 0)),
		pl.BlockSpec((block_k_major, head_dim), lambda *_: (0, 0)),
	]

	kernel = functools.partial(
		_flash_attention_dkv_kernel,
		block_q=block_q,
		block_k=block_k,
		sm_scale=sm_scale,
		causal=causal,
		mask_value=mask_value,
		q_seq_len=q_seq_len,
	)
	name_scope = (
		f"flash_mha_bwd_dkv_{block_q_major=}_{block_q=}_{block_k_major=}_{block_k=}"
	)
	with jax.named_scope(name_scope):
		dk, dv, _, _ = pl.pallas_call(
			kernel,
			out_shape=out_shapes,
			grid_spec=pltpu.PrefetchScalarGridSpec(
				num_scalar_prefetch=2, in_specs=in_specs, out_specs=out_specs, grid=grid
			),
			debug=debug,
			mosaic_params=dict(
				dimension_semantics=(
					"parallel",
					"parallel",
					"parallel",
					"arbitrary",
				)
			),
		)(
			q_chunk_idx_start,
			k_chunk_idx_start,
			q,
			k,
			v,
			ab,
			q_segment_ids,
			kv_segment_ids,
			l,
			m,
			do,
			di,
		)
		assert dk.shape == k.shape
		assert dv.shape == v.shape
	return dk, dv


def _flash_attention_dq_kernel(
	q_chunk_idx_start_ref,
	k_chunk_idx_start_ref,
	q_tile_ref,
	k_tile_ref,
	v_tile_ref,
	ab_tile_ref,
	q_segment_ids_tile_ref,
	kv_segment_ids_tile_ref,
	l_tile_ref,
	m_tile_ref,
	do_tile_ref,
	di_tile_ref,
	dq_tile_ref,
	dq_scratch_ref,
	ds_tile_ref,
	*,
	sm_scale: float,
	causal: bool,
	mask_value: float,
	kv_seq_len: int,
	block_k: int,
):
	_, _, block_k_major, _ = k_tile_ref.shape
	_, _, block_q_major, _ = q_tile_ref.shape

	kv_seq_index = pl.program_id(axis=3)
	q_seq_index = pl.program_id(axis=2)

	q_chunk_idx_start = q_chunk_idx_start_ref[0]
	k_chunk_idx_start = k_chunk_idx_start_ref[0]

	@pl.when(kv_seq_index == 0)
	def start_new_sequence():
		dq_scratch_ref[:, :] = jnp.zeros(dq_scratch_ref.shape, dq_scratch_ref.dtype)

	def body(i, _):
		k_slice = pl.ds(i * block_k, block_k)
		q = q_tile_ref[0, 0, :, :]
		k = pl.load(
			k_tile_ref,
			(0, 0, k_slice, slice(None)),
		)  # [block_k, head_dim]
		v = pl.load(
			v_tile_ref,
			(0, 0, k_slice, slice(None)),
		)  # [block_k, head_dim]
		l = l_tile_ref[0, 0, :, :]  # [block_q_major, 128]
		m = m_tile_ref[0, 0, :, :]  # [block_q_major, 128]
		do = do_tile_ref[0, 0, :, :]  # [block_q_major, head_dim]
		di = di_tile_ref[0, 0, :].astype(jnp.float32)  # [block_q_major, 128]

		capped_logits = jax.lax.dot_general(
			q, k, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
		)

		if ab_tile_ref is not None:
			ab = pl.load(
				ab_tile_ref,
				(0, pl.dslice(0, block_q_major), pl.dslice(i * block_k, block_k)),
			).astype(jnp.float32)
			capped_logits += ab

		if sm_scale != 1.0:
			capped_logits *= sm_scale

		mask = None
		if q_segment_ids_tile_ref is not None:
			repeats, rem = divmod(block_k, NUM_LANES)
			if rem:
				raise NotImplementedError(f"kv block size must be a multiple of {NUM_LANES}")
			q_segment_ids = pltpu.repeat(
				q_segment_ids_tile_ref[0], repeats, axis=1
			)  # [block_q, block_k].
			kv_segment_ids = pl.load(
				kv_segment_ids_tile_ref, (slice(None), 0, k_slice)
			)  # [1, block_k].
			mask = jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)

		if causal:
			mask_shape = (block_q_major, block_k)
			row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
			row_ids += (q_seq_index + q_chunk_idx_start) * block_q_major
			col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
			col_ids += (kv_seq_index + k_chunk_idx_start) * block_k_major + i * block_k
			causal_mask = col_ids <= row_ids
			mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
		capped_logits = (
			capped_logits
			if mask is None
			else capped_logits + jnp.where(mask, 0.0, mask_value)
		)

		p = jnp.exp(capped_logits - pltpu.repeat(m, block_k // MIN_BLOCK_SIZE, axis=1))
		p = p * pltpu.repeat(
			1 / l, block_k // MIN_BLOCK_SIZE, axis=1
		)  # [block_q_major, block_k]

		# di: [block_q_major, 128]
		# do: [block_q_major, head_dim]
		# v: [block_k_major, head_dim]
		dp = jax.lax.dot_general(
			do,
			v,
			TRANS_B_DIM_NUMBERS,
			preferred_element_type=jnp.float32,
		)
		ds = (dp - pltpu.repeat(di, block_k // MIN_BLOCK_SIZE, axis=1)) * p

		if sm_scale != 1.0:
			ds = ds * sm_scale

		if ds_tile_ref is not None:
			pl.store(
				ds_tile_ref,
				(0, pl.dslice(None), pl.dslice(i * block_k, block_k)),
				ds.astype(ds_tile_ref.dtype),
			)

		# dp: [block_q_major, block_k]
		# k: [block_k, head_dim]
		dq_scratch_ref[:, :] += lax.dot(
			ds.astype(k.dtype),
			k,
			preferred_element_type=jnp.float32,
		).astype(dq_scratch_ref.dtype)

	if causal:
		should_run = below_or_on_diag(
			q_seq_index + q_chunk_idx_start,
			block_q_major,
			kv_seq_index + k_chunk_idx_start,
			block_k_major,
		)
		should_not_run = lax.select(should_run, False, True)
	else:
		should_run = True
		should_not_run = False  # type: ignore

	@pl.when(should_run)
	def run():
		lax.fori_loop(0, block_k_major // block_k, body, None, unroll=True)

	@pl.when(should_not_run)
	def zero_out_ds():
		if ds_tile_ref is not None:
			ds_tile_ref[...] = jnp.zeros_like(ds_tile_ref)

	@pl.when(kv_seq_index == kv_seq_len // block_k_major - 1)
	def end_of_kv_sequence():
		dq_tile_ref[0, 0, :, :] = dq_scratch_ref[...].astype(dq_tile_ref)
		dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)


def _flash_attention_bwd_dq(
	q_chunk_idx_start,
	k_chunk_idx_start,
	q,
	k,
	v,
	ab,
	segment_ids,
	l,
	m,
	do,
	di,
	*,
	block_q_major: Optional[int],
	block_k_major: Optional[int],
	block_k: Optional[int],
	sm_scale: float,
	causal: bool,
	mask_value: float,
	debug: bool,
):
	batch_size, num_heads, q_seq_len, head_dim = q.shape
	_, _, kv_seq_len, _ = k.shape
	q_chunk_idx_start, k_chunk_idx_start = (
		q_chunk_idx_start[None],
		k_chunk_idx_start[None],
	)
	_verify_block("block_q_dq", "q_seq_len", block_q_major, q_seq_len)
	_verify_block("block_k_major_dq", "kv_seq_len", block_k_major, kv_seq_len)
	_verify_block("block_k_dq", "block_k", block_k, kv_seq_len)

	# Broadcast out scalar values
	m = jnp.broadcast_to(m[..., None], (*m.shape, MIN_BLOCK_SIZE))
	l = jnp.broadcast_to(l[..., None], (*l.shape, MIN_BLOCK_SIZE))
	# Preprocess contraction for bwd pass
	di = jnp.broadcast_to(di[..., None], (*di.shape, block_k_major))

	grid = (
		batch_size,
		num_heads,
		q_seq_len // block_q_major,
		kv_seq_len // block_k_major,
	)

	def qo_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, q_seq_index, 0)

	qo_spec = pl.BlockSpec((1, 1, block_q_major, head_dim), qo_index_map)
	do_spec = qo_spec

	def kv_index_map(
		batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref
	):
		if causal:
			# If the kv block is skipped, prefetch the next valid kv block, i.e. the
			# 0th one to be used for the next block_q rows.
			next_kv_index = lax.select(
				below_or_on_diag(
					q_seq_index + q_idx_ref[0],
					block_q_major,
					kv_seq_index + k_idx_ref[0],
					block_k_major,
				),
				kv_seq_index,
				0,
			)
		else:
			next_kv_index = kv_seq_index
		return (batch_index, head_index, next_kv_index, 0)

	kv_spec = pl.BlockSpec((1, 1, block_k_major, head_dim), kv_index_map)
	assert kv_spec.block_shape is not None
	assert k.ndim == len(kv_spec.block_shape)
	assert v.ndim == len(kv_spec.block_shape)

	def lm_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
		return (batch_index, head_index, q_seq_index, 0)

	lm_spec = pl.BlockSpec((1, 1, block_q_major, MIN_BLOCK_SIZE), lm_index_map)
	assert lm_spec.block_shape is not None
	assert l.ndim == len(lm_spec.block_shape)
	assert m.ndim == len(lm_spec.block_shape)

	di_spec = pl.BlockSpec((1, 1, block_q_major, MIN_BLOCK_SIZE), qo_index_map)
	assert di_spec.block_shape is not None
	assert di.ndim == len(di_spec.block_shape)

	def ab_index_map(
		batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref
	):
		return (batch_index, 0, kv_seq_index)

	if ab is not None:
		ab = ab[:, None].repeat(block_q_major, axis=1)

	dab_spec = (
		pl.BlockSpec((1, block_q_major, block_k_major), ab_index_map)
		if ab is not None
		else None
	)

	q_segment_ids_spec = kv_segment_ids_spec = None
	q_segment_ids = kv_segment_ids = None
	if segment_ids is not None:

		def q_segment_ids_index_map(
			batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref
		):
			del head_index
			return (batch_index, q_seq_index, 0)

		def kv_segment_ids_index_map(
			batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref
		):
			del head_index
			if causal:
				# If the kv block is skipped, prefetch the next valid kv block, i.e. the
				# 0th one to be used for the next block_q rows.
				next_kv_index = lax.select(
					below_or_on_diag(
						q_seq_index + q_idx_ref[0],
						block_q_major,
						kv_seq_index + k_idx_ref[0],
						block_k_major,
					),
					kv_seq_index,
					0,
				)
			else:
				next_kv_index = kv_seq_index
			return (batch_index, 0, next_kv_index)

		q_segment_ids_spec = pl.BlockSpec(
			(1, block_q_major, NUM_LANES),
			q_segment_ids_index_map,
		)
		kv_segment_ids_spec = pl.BlockSpec(
			(1, NUM_SUBLANES, block_k_major),
			kv_segment_ids_index_map,
		)

		q_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.q,
			(batch_size, q_seq_len, NUM_LANES),
			(
				0,
				1,
			),
		)
		kv_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.kv,
			(batch_size, NUM_SUBLANES, kv_seq_len),
			(
				0,
				2,
			),
		)

	in_specs = [
		qo_spec,
		kv_spec,
		kv_spec,
		dab_spec,
		q_segment_ids_spec,
		kv_segment_ids_spec,
		lm_spec,
		lm_spec,
		do_spec,
		di_spec,
	]

	out_shapes = [
		jax.ShapeDtypeStruct(q.shape, q.dtype),
		jax.ShapeDtypeStruct((block_q_major, head_dim), jnp.float32),
		jax.ShapeDtypeStruct(ab.shape, ab.dtype) if ab is not None else None,
	]
	dq_spec = pl.BlockSpec((1, 1, block_q_major, head_dim), qo_index_map)
	out_specs = [
		dq_spec,
		pl.BlockSpec((block_q_major, head_dim), lambda *_: (0, 0)),
		dab_spec,
	]

	kernel = functools.partial(
		_flash_attention_dq_kernel,
		sm_scale=sm_scale,
		causal=causal,
		mask_value=mask_value,
		block_k=block_k,
		kv_seq_len=kv_seq_len,
	)
	name_scope = f"flash_mha_bwd_dq_{block_q_major=}_{block_k_major=}_{block_k=}"
	with jax.named_scope(name_scope):
		dq, _, ds = pl.pallas_call(
			kernel,
			out_shape=out_shapes,
			grid_spec=pltpu.PrefetchScalarGridSpec(
				num_scalar_prefetch=2, in_specs=in_specs, out_specs=out_specs, grid=grid
			),
			debug=debug,
			mosaic_params=dict(
				dimension_semantics=(
					"parallel",
					"parallel",
					"parallel",
					"arbitrary",
				)
			),
		)(
			q_chunk_idx_start,
			k_chunk_idx_start,
			q,
			k,
			v,
			ab,
			q_segment_ids,
			kv_segment_ids,
			l,
			m,
			do,
			di,
		)

	return dq, ds


def _verify_block(block_name, dim_name, block, dim, should_divide=True):
	if block > dim:
		raise ValueError(
			f"{block_name}={block} should be smaller or equal to {dim_name}={dim}"
		)
	if should_divide and dim % block != 0:
		raise ValueError(f"{dim_name}={dim} should be divisible by {block_name}={block}")



# jax.experimental.pallas.ops.gpu
from __future__ import annotations

import functools

import jax
import jax.numpy as jnp
from jax import lax
from jax._src.lax.control_flow.for_loop import for_loop
from jax.experimental import pallas as pl


def rms_norm_forward_kernel(
	x_ref,
	weight_ref,
	bias_ref,  # Input arrays
	o_ref,
	rstd_ref=None,  # Output arrays
	*,
	eps: float,
	block_size: int,
):
	n_col = x_ref.shape[0]

	def var_body(i, acc_ref):
		col_idx = i * block_size + jnp.arange(block_size)
		mask = col_idx < n_col
		a = pl.load(
			x_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		a = jnp.where(mask, a, 0.0)
		acc_ref[:] += a * a

	var = (
		for_loop(pl.cdiv(n_col, block_size), var_body, jnp.zeros(block_size)).sum() / n_col
	)
	rstd = 1 / jnp.sqrt(var + eps)
	if rstd_ref is not None:
		rstd_ref[...] = rstd.astype(rstd_ref.dtype)

	def body(i, _):
		col_idx = i * block_size + jnp.arange(block_size)
		mask = col_idx < n_col
		weight = pl.load(weight_ref, (col_idx,), mask=mask)
		bias = pl.load(bias_ref, (col_idx,), mask=mask)
		x = pl.load(
			x_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_first"
		).astype(jnp.float32)
		out = x * rstd * weight + bias
		pl.store(o_ref, (col_idx,), out.astype(o_ref.dtype), mask=mask)

	for_loop(pl.cdiv(n_col, block_size), body, ())


def rms_norm_forward(
	x,
	weight,
	bias,
	num_warps: int | None = None,
	num_stages: int | None = 3,
	eps: float = 1e-5,
	backward_pass_impl: str = "triton",
	interpret: bool = False,
):
	del num_stages
	del backward_pass_impl
	n = x.shape[-1]
	# Triton heuristics
	# Less than 64KB per feature: enqueue fused kernel
	max_fused_size = 65536 // x.dtype.itemsize
	block_size = min(max_fused_size, pl.next_power_of_2(n))
	block_size = min(max(block_size, 128), 4096)
	num_warps = min(max(block_size // 256, 1), 8)

	kernel = functools.partial(rms_norm_forward_kernel, eps=eps, block_size=block_size)
	out_shape = [
		jax.ShapeDtypeStruct(shape=(n,), dtype=x.dtype),
		jax.ShapeDtypeStruct(shape=(), dtype=x.dtype),
	]
	method = pl.pallas_call(
		kernel,
		compiler_params=dict(triton=dict(num_warps=num_warps)),
		grid=(),
		out_shape=out_shape,
		debug=False,
		interpret=interpret,
		name="rms_forward",
	)

	method = jax.vmap(jax.vmap(method, in_axes=(0, None, None)), in_axes=(0, None, None))
	out, rstd = method(x, weight, bias)
	return out, (x, weight, bias, rstd)


def rms_norm_backward_kernel_dx(
	# Inputs
	x_ref,
	weight_ref,
	bias_ref,
	do_ref,
	rstd_ref,
	# Outputs
	dx_ref,
	*,
	eps: float,
	block_size: int,
):
	n_col = x_ref.shape[0]

	def mean_body(i, c1_acc_ref):
		col_idx = i * block_size + jnp.arange(block_size)
		mask = col_idx < n_col
		a = pl.load(
			x_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		dout = pl.load(
			do_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		weight = pl.load(
			weight_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		a_hat = a * rstd_ref[...]
		wdout = weight * dout
		c1_acc_ref[:] += a_hat * wdout

	c1 = for_loop(pl.cdiv(n_col, block_size), mean_body, jnp.zeros(block_size))
	c1 = c1.sum() / n_col

	def dx_body(i, acc_ref):
		col_idx = i * block_size + jnp.arange(block_size)
		mask = col_idx < n_col
		a = pl.load(
			x_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		dout = pl.load(
			do_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		weight = pl.load(
			weight_ref, (col_idx,), mask=mask, other=0.0, eviction_policy="evict_last"
		).astype(jnp.float32)
		a_hat = a * rstd_ref[...]
		wdout = weight * dout
		da = (wdout - (a_hat * c1)) * rstd_ref[...]
		pl.store(dx_ref, (col_idx,), da.astype(dx_ref.dtype), mask=mask)

	for_loop(pl.cdiv(n_col, block_size), dx_body, ())


def rms_norm_backward_kernel_dw_db(
	# Inputs
	x_ref,
	weight_ref,
	bias_ref,
	do_ref,
	rstd_ref,
	# Outputs
	dw_ref,
	db_ref,
	*,
	eps: float,
	block_m: int,
	block_n: int,
):
	m, n_col = x_ref.shape
	j = pl.program_id(0)
	col_idx = j * block_n + jnp.arange(block_n)
	col_mask = col_idx < n_col

	def body(i, acc_ref):
		row_idx = i * block_m + jnp.arange(block_m)
		row_mask = row_idx < m
		mask = row_mask[:, None] & col_mask[None, :]
		a = pl.load(x_ref, (row_idx[:, None], col_idx[None]), mask=mask, other=0.0).astype(
			jnp.float32
		)
		dout = pl.load(
			do_ref, (row_idx[:, None], col_idx[None]), mask=mask, other=0.0
		).astype(jnp.float32)
		rstd = pl.load(rstd_ref, (row_idx,), mask=row_mask, other=0.0).astype(jnp.float32)
		a_hat = a * rstd[:, None]
		dw_acc_ref, db_acc_ref = acc_ref
		dw_acc_ref[:] += (dout * a_hat).sum(axis=0)
		db_acc_ref[:] += dout.sum(axis=0)

	dw_acc, db_acc = for_loop(
		pl.cdiv(m, block_m), body, (jnp.zeros(block_n), jnp.zeros(block_n))
	)
	pl.store(dw_ref, (col_idx,), dw_acc.astype(dw_ref.dtype), mask=col_mask)
	pl.store(db_ref, (col_idx,), db_acc.astype(db_ref.dtype), mask=col_mask)


def rms_norm_backward(
	num_warps: int | None,
	num_stages: int | None,
	eps: float,
	backward_pass_impl: str,
	interpret: bool,
	res,
	do,
):
	del num_stages
	x, weight, bias, rstd = res
	if backward_pass_impl == "xla":
		return jax.vjp(rms_norm_reference, x, weight, bias)[1](do)

	*shape_prefix, n = x.shape
	reshaped_x = x.reshape((-1, n))
	reshaped_rstd = rstd.reshape((-1,))
	reshaped_do = do.reshape((-1, n))
	# Triton heuristics
	# Less than 64KB per feature: enqueue fused kernel
	max_fused_size = 65536 // x.dtype.itemsize
	block_size = min(max_fused_size, pl.next_power_of_2(n))
	block_size = min(max(block_size, 128), 4096)
	num_warps = min(max(block_size // 256, 1), 8)

	# rms_norm_backward_kernel_dx parallel over batch dims
	kernel = functools.partial(
		rms_norm_backward_kernel_dx, eps=eps, block_size=block_size
	)
	out_shape_dx = jax.ShapeDtypeStruct(shape=(n,), dtype=x.dtype)
	method = pl.pallas_call(
		kernel,
		compiler_params=dict(triton=dict(num_warps=num_warps)),
		grid=(),
		out_shape=out_shape_dx,
		debug=False,
		interpret=interpret,
		name="ln_backward_dx",
	)

	method = jax.vmap(method, in_axes=(0, None, None, 0, 0))
	dx = method(reshaped_x, weight, bias, reshaped_do, reshaped_rstd)
	dx = dx.reshape((*shape_prefix, n))

	# rms_norm_backward_kernel_dw_db reduce over batch dims
	# Triton heuristics
	if n > 10240:
		block_n = 128
		block_m = 32
		num_warps = 4
	else:
		# maximize occupancy for small N
		block_n = 16
		block_m = 16
		num_warps = 8
	kernel = functools.partial(
		rms_norm_backward_kernel_dw_db, eps=eps, block_m=block_m, block_n=block_n
	)
	out_shape_dwbias = [
		jax.ShapeDtypeStruct(shape=weight.shape, dtype=weight.dtype),
		jax.ShapeDtypeStruct(shape=bias.shape, dtype=bias.dtype),
	]
	grid_ = (pl.cdiv(reshaped_x.shape[1], block_n),)
	method = pl.pallas_call(
		kernel,
		compiler_params=dict(triton=dict(num_warps=num_warps)),
		grid=grid_,
		out_shape=out_shape_dwbias,
		debug=False,
		interpret=interpret,
		name="ln_backward_dw_db",
	)
	dw, dbias = method(reshaped_x, weight, bias, reshaped_do, reshaped_rstd)
	return dx, dw, dbias


@functools.partial(jax.custom_vjp, nondiff_argnums=[3, 4, 5, 6, 7])
@functools.partial(
	jax.jit,
	static_argnames=[
		"num_warps",
		"num_stages",
		"num_stages",
		"eps",
		"backward_pass_impl",
		"interpret",
	],
)
def rms_norm(
	x,
	weight,
	bias,
	num_warps: int | None = None,
	num_stages: int | None = 3,
	eps: float = 1e-5,
	backward_pass_impl: str = "triton",
	interpret: bool = False,
):
	n = x.shape[-1]
	# Triton heuristics
	# Less than 64KB per feature: enqueue fused kernel
	max_fused_size = 65536 // x.dtype.itemsize
	block_size = min(max_fused_size, pl.next_power_of_2(n))
	block_size = min(max(block_size, 128), 4096)
	num_warps = min(max(block_size // 256, 1), 8)

	kernel = functools.partial(rms_norm_forward_kernel, eps=eps, block_size=block_size)
	out_shape = jax.ShapeDtypeStruct(shape=(n,), dtype=x.dtype)
	method = pl.pallas_call(
		kernel,
		compiler_params=dict(triton=dict(num_warps=num_warps, num_stages=num_stages)),
		grid=(),
		out_shape=out_shape,
		debug=False,
		interpret=interpret,
	)
	method = jax.vmap(jax.vmap(method, in_axes=(0, None, None)), in_axes=(0, None, None))
	return method(x, weight, bias)


rms_norm.defvjp(rms_norm_forward, rms_norm_backward)


@functools.partial(jax.jit, static_argnames=["eps"])
@functools.partial(jax.vmap, in_axes=(0, None, None), out_axes=0)
def rms_norm_reference(x, weight, bias, *, eps: float = 1e-5):
	var = jnp.mean(jnp.square(x), axis=1)
	mul = lax.rsqrt(var + eps)
	return x * mul[:, None] * weight[None] + bias[None]



import os
from functools import partial
from typing import Callable, Literal, Optional, Tuple, Union

import flax
import jax
import msgpack
import safetensors.flax
import tqdm
from flax import struct
from flax.serialization import from_bytes, from_state_dict, to_bytes, to_state_dict
from flax.traverse_util import empty_node, flatten_dict, unflatten_dict
from jax import numpy as jnp
from fjformer.utils import get_logger

logger = get_logger(__name__)


def process_tensor(key, shard_fns, mismatch_allowed, manager):
	tensor = manager.get_tensor(key)
	if shard_fns is not None:
		try:
			callable_func = shard_fns.get(key)
			if callable_func is None:
				if not mismatch_allowed:
					raise KeyError(
						f"Shard Function {key} is None and NoneType OBJ is not callable."
					)
				return key, tensor, 1
			return key, callable_func(tensor), 0
		except KeyError as k_err:
			if mismatch_allowed:
				return key, tensor, 1
			raise KeyError(k_err) from None
	return key, tensor, 0


def is_flatten(pytree: Union[dict, struct.PyTreeNode]) -> bool:
	"""
	Check if the given PyTree is flattened.

	Args:
	    pytree: The PyTree to check.

	Returns:
	    True if the PyTree is flattened, False otherwise.
	"""
	return True if isinstance([k for k in pytree.keys()][0], tuple) else False


def get_dtype(tensor: jax.Array, dtype: Optional[Union[str, jnp.dtype]]) -> jax.Array:
	"""
	Get the tensor with the specified data type.

	Args:
	    tensor: The input tensor.
	    dtype: The desired data type.

	Returns:
	    The tensor with the specified data type.
	"""
	if dtype is None or dtype == "":
		return tensor
	if isinstance(dtype, str):
		dtype = {
			"bf16": jnp.bfloat16,
			"bfloat16": jnp.bfloat16,
			"fp16": jnp.float16,
			"float16": jnp.float16,
			"fp32": jnp.float32,
			"float32": jnp.float32,
			"fp64": jnp.float64,
			"float64": jnp.float64,
		}[dtype]
	float_dtypes = (jnp.bfloat16, jnp.float16, jnp.float32, jnp.float64)
	if getattr(tensor, "dtype", None) in float_dtypes:
		tensor = tensor.astype(dtype)
	return tensor


class CheckpointManager(object):
	"""
	A class to manage saving and loading checkpoints.

	Args:
	    checkpoint_dir: The directory to save checkpoints to.
	    enable: Whether to enable saving and loading checkpoints.
	    float_dtype: The floating-point data type to use for saving checkpoints.
	    save_optimizer_state: Whether to save the optimizer state in the checkpoint.
	    verbose: Whether to print verbose output.
	"""

	def __init__(
		self,
		checkpoint_dir: Union[str, os.PathLike],
		enable: bool = True,
		float_dtype: Union[str, jnp.dtype] = "bf16",
		save_optimizer_state: bool = True,
		verbose: bool = False,
	):
		self.float_dtype = float_dtype
		self.save_optimizer_state = save_optimizer_state
		self.checkpoint_dir = checkpoint_dir
		self.enable = enable
		self.verbose = verbose

	def save_checkpoint(
		self,
		state: struct.PyTreeNode,
		filename: Union[str, os.PathLike],
		gather_fns: Optional[dict[Callable]] = None,
		mismatch_allowed: bool = True,
	):
		"""
		Save a checkpoint to the given filename.

		Args:
		    state: The state dictionary to save.
		    filename: The filename to save the checkpoint to.
		    gather_fns: A dictionary of functions to gather the state before saving.
		    mismatch_allowed: Whether to allow mismatches between the state dictionary and gather functions.
		"""
		if self.enable:
			path = os.path.join(self.checkpoint_dir, filename)
		else:
			path = "/dev/null"
		self.save_state_to_file(
			state, path, gather_fns, self.float_dtype, mismatch_allowed=mismatch_allowed
		)

	@staticmethod
	def load_checkpoint_safe(
		path: Union[str, os.PathLike],
		target: Optional[struct.PyTreeNode] = None,
		shard_fns: Optional[dict[Callable]] = None,
		verbose: bool = False,
		mismatch_allowed: bool = True,
	) -> Tuple[Union[struct.PyTreeNode, dict], dict]:
		"""
		Load a checkpoint from the given path.

		Args:
		                path: The path to the checkpoint file.
		                target: The target PyTree to load the checkpoint into.
		                shard_fns: A dictionary of functions to shard the state after loading.
		                verbose: Whether to print verbose output.
		                mismatch_allowed: Whether to allow mismatches between the state dictionary and shard functions.

		Returns:
		                A tuple containing the loaded state dictionary and metadata.
		"""
		with safetensors.safe_open(path, framework="flax") as f:
			metadata = f.metadata()
			keys = list(f.keys())

			if shard_fns is not None:
				if not is_flatten(shard_fns):
					shard_fns = flatten_dict(shard_fns, sep=".")

			process_func = partial(
				process_tensor,
				shard_fns=shard_fns,
				mismatch_allowed=mismatch_allowed,
				manager=f,
			)
			results = [
				process_func(key)
				for key in tqdm.tqdm(
					keys,
					desc="Loading",
					total=len(keys),
				)
			]

		state = {key: tensor for key, tensor, _ in results}
		shard_functions_mismatch = sum(mismatch for _, _, mismatch in results)

		if verbose and shard_functions_mismatch != 0:
			logger.info(f"Sharding mismatch: {shard_functions_mismatch}")

		if target is not None:
			flattened_target = flatten_dict(to_state_dict(target), keep_empty_nodes=True)
			for key, value in flattened_target.items():
				if key not in state and value == empty_node:
					state[key] = value

		state = unflatten_dict(state, sep=".")
		if target is None:
			return state, metadata

		return from_state_dict(target, state), metadata

	@staticmethod
	def save_checkpoint_safe(
		state: struct.PyTreeNode,
		path: Union[str, os.PathLike],
		gather_fns: Optional[dict[Callable]] = None,
		float_dtype: Optional[Union[str, jnp.dtype]] = None,
		verbose: bool = True,
		mismatch_allowed: bool = True,
		metadata: Optional[dict[str, str]] = None,
	):
		"""
		Save a checkpoint to the given path using SafeTensors.

		Args:
		    state: The state dictionary to save.
		    path: The path to the checkpoint file.
		    gather_fns: A dictionary of functions to gather the state before saving.
		    float_dtype: The floating-point data type to use for saving the checkpoint.
		    verbose: Whether to print verbose output.
		    mismatch_allowed: Whether to allow mismatches between the state dictionary and gather functions.
		    metadata: Additional metadata to store in the checkpoint.
		"""
		state = to_state_dict(state)
		gather_functions_mismatch = 0
		if is_flatten(state):
			state = unflatten_dict(state)

		if gather_fns is not None:
			# Example:
			# gather_fns = {"params": {"Dense_0": lambda x: x[0]}}

			if not is_flatten(gather_fns):
				gather_fns = flatten_dict(gather_fns)
			state = flatten_dict(state)
			pbar_gather = tqdm.tqdm(
				list(state.keys()), desc="Gathering State", disable=not verbose
			)
			for key in pbar_gather:
				try:
					callable_func = gather_fns[key]
					if callable_func is None and not mismatch_allowed:
						raise KeyError(
							f"Gather Function {key} is None and NoneType OBJ is not callable."
						)
					if callable_func is None:
						gather_functions_mismatch += 1
					else:
						state[key] = callable_func(state[key])
				except KeyError as e:
					if mismatch_allowed:
						pbar_gather.set_postfix(gather_mismatch=gather_functions_mismatch)
					else:
						raise KeyError(e) from None
				pbar_gather.update(1)
		state = flax.traverse_util.flatten_dict(state, sep=".")
		for key in list(state.keys()):
			if not isinstance(state[key], jax.Array):
				state[key] = jnp.array(state[key])
			state[key] = get_dtype(state[key], float_dtype)

		safetensors.flax.save_file(tensors=state, filename=path, metadata=metadata)

	@staticmethod
	def save_state_to_file(
		state: struct.PyTreeNode,
		path: Union[str, os.PathLike],
		gather_fns: Optional[dict[Callable]] = None,
		float_dtype: Optional[Union[str, jnp.dtype]] = None,
		verbose: bool = False,
		mismatch_allowed: bool = True,
	):
		"""
		Save the state dictionary to a file.

		Args:
		    state: The state dictionary to save.
		    path: The path to the file to save the state dictionary to.
		    gather_fns: A dictionary of functions to gather the state before saving.
		    float_dtype: The floating-point data type to use for saving the state dictionary.
		    verbose: Whether to print verbose output.
		    mismatch_allowed: Whether to allow mismatches between the state dictionary and gather functions.
		"""
		state = to_state_dict(state)
		packer = msgpack.Packer()
		flatten_state = flatten_dict(state)
		if gather_fns is not None:
			gather_fns = flatten_dict(to_state_dict(gather_fns))
		pbar = tqdm.tqdm(
			flatten_state.items(),
			disable=not verbose,
			desc="Saving State to File",
		)

		gather_functions_mismatch = 0

		with open(path, "wb") as stream:
			for key, value in pbar:
				if gather_fns is not None:
					try:
						callable_func = gather_fns[key]
						if callable_func is None and not mismatch_allowed:
							raise KeyError(
								f"Gather Function {key} is None and NoneType OBJ is not callable."
							)
						value = callable_func(value) if callable_func is not None else value
						if callable_func is None:
							gather_functions_mismatch += 1
					except KeyError as k_err:
						if mismatch_allowed:
							gather_functions_mismatch += 1
						else:
							raise KeyError(k_err) from None
				pbar.set_postfix(gather_functions_mismatch=gather_functions_mismatch)
				value = get_dtype(value, float_dtype)
				stream.write(packer.pack((key, to_bytes(value))))

	def save_pickle(self, obj: object, filename: Union[str, os.PathLike]):
		"""
		Save an object to a pickle file.

		Args:
		    obj: The object to save.
		    filename: The filename to save the object to.
		"""
		import pickle

		def save_pickle(obj_, path_):
			with open(path_, "wb") as stream:
				pickle.dump(obj_, stream)

		if self.enable:
			path = os.path.join(self.checkpoint_dir, filename)
		else:
			path = "/dev/null"
		save_pickle(obj, path)

	def save_all(
		self,
		state: struct.PyTreeNode,
		gather_fns: dict[Callable],
		metadata: Optional[dict] = None,
		dataset: Optional[object] = None,
		milestone: bool = False,
	):
		"""
		Save all components of a checkpoint.

		Args:
		    state: The state dictionary to save.
		    gather_fns: A dictionary of functions to gather the state before saving.
		    metadata: Additional metadata to save.
		    dataset: The dataset to save.
		    milestone: Whether to save a milestone checkpoint.
		"""
		step = int(jax.device_get(state.step))
		if self.save_optimizer_state:
			checkpoint_state = state
			checkpoint_name = "streaming_state"
			checkpoint_gather_fns = gather_fns
		else:
			checkpoint_state = state.params["params"]
			checkpoint_name = "streaming_params"
			checkpoint_gather_fns = gather_fns.params["params"]

		if milestone:
			# Save a milestone checkpoint that will not be overwritten
			self.save_pickle(metadata, f"metadata_{step}.pkl")
			self.save_pickle(dataset, f"dataset_{step}.pkl")
			self.save_checkpoint(
				checkpoint_state, f"{checkpoint_name}_{step}", checkpoint_gather_fns
			)
		else:
			# Save a normal checkpoint that can be overwritten
			self.save_pickle(metadata, "metadata.pkl")
			self.save_pickle(dataset, "dataset.pkl")
			self.save_checkpoint(
				checkpoint_state, f"{checkpoint_name}", checkpoint_gather_fns
			)

	@staticmethod
	def load_checkpoint(
		path: Union[str, os.PathLike],
		target: Optional[struct.PyTreeNode] = None,
		shard_fns: Optional[dict[Callable]] = None,
		remove_dict_prefix: Optional[Tuple[str, ...]] = None,
		verbose: bool = False,
		mismatch_allowed: bool = True,
	) -> Union[struct.PyTreeNode, dict]:
		"""
		Load a checkpoint from the given path.

		Args:
		    path: The path to the checkpoint file.
		    target: The target PyTree to load the checkpoint into.
		    shard_fns: A dictionary of functions to shard the state after loading.
		    remove_dict_prefix: A tuple of strings representing the prefix to remove from the state dictionary keys.
		    verbose: Whether to print verbose output.
		    mismatch_allowed: Whether to allow mismatches between the state dictionary and shard functions.

		Returns:
		    The loaded state dictionary.
		"""
		if shard_fns is not None:
			shard_fns = flatten_dict(to_state_dict(shard_fns))
		if remove_dict_prefix is not None:
			remove_dict_prefix = tuple(remove_dict_prefix)
		flatten_state = {}

		shard_functions_mismatch = 0
		with open(path, "rb") as fin:
			unpacker = msgpack.Unpacker(fin, read_size=83886080, max_buffer_size=0)
			pbar = tqdm.tqdm(
				unpacker, disable=not verbose, desc="Loading Checkpoints From File"
			)
			for key, value in pbar:
				key = tuple(key)
				if remove_dict_prefix is not None:
					if key[: len(remove_dict_prefix)] == remove_dict_prefix:
						key = key[len(remove_dict_prefix) :]
					else:
						continue

				tensor = from_bytes(None, value)
				if shard_fns is not None:
					try:
						callable_func = shard_fns[key]
						if callable_func is None and not mismatch_allowed:
							raise KeyError(
								f"Shard Function {key} is None and NoneType OBJ is not callable."
							)
						tensor = callable_func(tensor) if callable_func is not None else tensor
						if callable_func is None:
							shard_functions_mismatch += 1
					except KeyError as k_err:
						if mismatch_allowed:
							shard_functions_mismatch += 1
						else:
							raise KeyError(k_err) from None
				flatten_state[key] = tensor
				pbar.set_postfix(shard_functions_mismatch=shard_functions_mismatch)
		if target is not None:
			flattened_target = flatten_dict(to_state_dict(target), keep_empty_nodes=True)
			for key, value in flattened_target.items():
				if key not in flatten_state and value == empty_node:
					flatten_state[key] = value

		state = unflatten_dict(flatten_state)
		if target is None:
			return state

		return from_state_dict(target, state)

	@staticmethod
	def load_flax_checkpoint(
		path: Union[str, os.PathLike],
		target: Optional[struct.PyTreeNode] = None,
		shard_fns: Optional[dict[Callable]] = None,
	) -> Union[struct.PyTreeNode, dict]:
		"""
		Load a standard flax checkpoint that's not saved with the
		msgpack streaming format.

		Args:
		    path: The path to the checkpoint file.
		    target: The target PyTree to load the checkpoint into.
		    shard_fns: A dictionary of functions to shard the state after loading.

		Returns:
		    The loaded state dictionary.
		"""
		with open(path, "rb") as fin:
			encoded_bytes = fin.read()

		state_dict = flax.serialization.msgpack_restore(encoded_bytes)
		if shard_fns is not None:
			shard_fns = to_state_dict(shard_fns)
			state_dict = jax.tree_util.tree_map(lambda fn, x: fn(x), shard_fns, state_dict)

		if target is None:
			return state_dict
		return from_state_dict(target, state_dict)

	@classmethod
	def load_state_checkpoint(
		cls,
		load_type: Literal["state", "state_params", "params", "flax_params"],
		load_path: Union[str, os.PathLike],
		state_target: Optional[struct.PyTreeNode] = None,
		state_shard_fns: Optional[dict[Callable]] = None,
		disallow_state: bool = False,
		mismatch_allowed: bool = True,
	) -> Tuple[Optional[struct.PyTreeNode], Optional[flax.core.frozen_dict.FrozenDict]]:
		"""
		Load a state checkpoint from the given path.

		Args:
		    load_type: The type of checkpoint to load.
		    load_path: The path to the checkpoint file.
		    state_target: The target PyTree to load the state into.
		    state_shard_fns: A dictionary of functions to shard the state after loading.
		    disallow_state: Whether to disallow loading the full state.
		    mismatch_allowed: Whether to allow mismatches between the state dictionary and shard functions.

		Returns:
		    A tuple containing the loaded state and parameters.
		"""
		if state_target is not None:
			params_target = state_target.params["params"]
		else:
			params_target = None

		if state_shard_fns is not None:
			params_shard_fns = state_shard_fns.params["params"]
		else:
			params_shard_fns = None

		if disallow_state:
			assert load_type != "state", "Loading full state is not allowed!"
		state = None
		restored_params = None
		if load_type == "state":
			state = cls.load_checkpoint(
				path=load_path,
				target=state_target,
				shard_fns=state_shard_fns,
				mismatch_allowed=mismatch_allowed,
			)
		elif load_type == "state_params":
			restored_params = cls.load_checkpoint(
				path=load_path,
				target=params_target,
				shard_fns=params_shard_fns,
				remove_dict_prefix=("params", "params"),
				mismatch_allowed=mismatch_allowed,
			)
			restored_params = flax.core.frozen_dict.freeze({"params": restored_params})
		elif load_type == "params":
			restored_params = cls.load_checkpoint(
				path=load_path,
				target=params_target,
				shard_fns=params_shard_fns,
				mismatch_allowed=mismatch_allowed,
			)
			restored_params = flax.core.frozen_dict.freeze({"params": restored_params})
		elif load_type == "flax_params":
			restored_params = cls.load_flax_checkpoint(
				path=load_path,
				target=params_target,
				shard_fns=params_shard_fns,
			)
			restored_params = flax.core.frozen_dict.freeze({"params": restored_params})
		else:
			raise ValueError(f"Invalid load_from type: {load_type}")

		return state, restored_params



import functools
from typing import NamedTuple

import chex
import jax
import jax.lax as lax
import jax.numpy as jnp
from einops import rearrange


class Carry(NamedTuple):
	numerator: chex.Array
	denominator: chex.Array
	max_so_far: chex.Array


def efficient_attention(
	query: chex.Array,
	key: chex.Array,
	value: chex.Array,
	bias: chex.Array = None,
	deterministic: bool = True,
	dropout_rng: chex.PRNGKey = None,
	attention_drop_rate: float = 0.0,
	causal: bool = True,
	query_chunk_size: int = 1024,
	key_chunk_size: int = 1024,
	dtype: chex.ArrayDType = jnp.float32,  # type:ignore
	policy=jax.checkpoint_policies.nothing_saveable(),  # noqa
	precision=None,
	float32_logits: bool = True,
	prevent_cse: bool = True,
):
	"""Memory-efficient attention implementation with optional chunking and causal masking.

	This function implements the multi-head scaled dot-product attention mechanism
	with several efficiency optimizations:

	- **Chunking:**
		- The inputs are split into chunks along the length dimension, and attention
		  is computed one chunk at a time.
		- This reduces memory usage significantly for long sequences, as it avoids
		  computing the entire attention matrix at once.
		- Chunking can be controlled using the `query_chunk_size` and `key_chunk_size`
		  parameters.
	- **Causal masking:**
		- For autoregressive tasks, a causal mask can be applied to prevent the attention
		  mechanism from attending to future tokens.
		- This is enabled by setting the `causal` parameter to `True`.
	- **Dropouts:**
		- Attention dropout is applied to the attention weights before softmax
		  normalization.
		- This can help to prevent overfitting and improve generalization performance.
	- **Checkpoint gradients:**
		- Gradient checkpointing can be used to further reduce memory usage during
		  training.
		- This is enabled by passing a checkpointing policy to the `policy` parameter.
	- **Mixed precision:**
		- The attention logits can be computed in float32 precision, even when the inputs
		  are in a lower precision (e.g., bfloat16).
		- This can improve numerical stability and reduce memory usage.

	Args:
		query: Query matrix with shape `(batch, q_len, num_heads, dim_per_head)`.
		key: Key matrix with shape `(batch, kv_len, num_heads, dim_per_head)`.
		value: Value matrix with shape `(batch, kv_len, num_heads, dim_per_head)`.
		bias: Optional bias matrix with shape broadcastable to `(batch, num_heads, q_len, kv_len)`.
		deterministic: Whether to apply dropout or not.
		dropout_rng: JAX PRNG key for dropout.
		attention_drop_rate: Dropout rate for attention weights.
		causal: Whether to apply a causal mask.
		query_chunk_size: Chunk size for the query matrix.
		key_chunk_size: Chunk size for the key and value matrices.
		dtype: Data type for the output.
		policy: Checkpointing policy for gradient checkpointing.
		precision: Precision for matrix multiplication.
		float32_logits: Whether to compute the attention logits in float32 precision.
		prevent_cse: Whether to prevent common subexpression elimination.

	Returns:
		The attended output with shape `(batch, q_len, num_heads, dim_per_head)`.

	Example usage:

	```python
	# Compute memory-efficient attention with causal masking
	output = efficient_attention(
	  query=query, key=key, value=value, causal=True, query_chunk_size=512
	)

	# Compute memory-efficient attention with dropout and gradient checkpointing
	output = efficient_attention(
	  query=query,
	  key=key,
	  value=value,
	  deterministic=False,
	  dropout_rng=rng,
	  attention_drop_rate=0.1,
	  policy=jax.checkpoint_policies.checkpoint_dots(),
	)
	```
	"""
	query = query / jnp.sqrt(query.shape[-1]).astype(dtype)
	if float32_logits:
		query = query.astype(jnp.float32)
		key = key.astype(jnp.float32)

	batch, q_len, num_heads, dim_per_head = query.shape
	batch, kv_len, kv_heads, dim_per_head = key.shape
	batch, kv_len, kv_heads, dim_per_head = value.shape

	num_q = q_len // query_chunk_size
	num_kv = kv_len // key_chunk_size
	query = query.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
	key = key.reshape((batch, num_kv, key_chunk_size, kv_heads, dim_per_head))
	value = value.reshape((batch, num_kv, key_chunk_size, kv_heads, dim_per_head))

	query = jnp.moveaxis(query, 1, 0)
	key = jnp.moveaxis(key, 1, 0)
	value = jnp.moveaxis(value, 1, 0)

	if bias is not None:
		for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):  # noqa
			assert bias_dim == 1 or bias_dim == broadcast_dim
	if not deterministic and attention_drop_rate > 0.0:
		attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
		attn_dropout = jax.random.bernoulli(
			attn_dropout_rng, attention_drop_rate, (batch, num_heads, q_len, kv_len)
		)
	else:
		attn_dropout = None

	_chunk_bias_fn = functools.partial(
		_chunk_attention_bias,
		query_chunk_size,
		key_chunk_size,
		bias,
		deterministic,
		attn_dropout,
		attention_drop_rate,
		causal,
		dtype,
	)

	def scan_attention(args):
		query_chunk, query_chunk_idx = args

		@functools.partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
		def scan_kv_block(carry, args):
			key_chunk, value_chunk, key_chunk_idx = args
			(numerator, denominator, prev_max_score) = carry
			attn_weights = jnp.einsum(
				"bqhd,bkhd->bqhk", query_chunk, key_chunk, precision=precision
			)
			bias_chunk = _chunk_bias_fn(query_chunk_idx, key_chunk_idx)
			bias_chunk = jnp.moveaxis(bias_chunk, 1, 2)
			attn_weights = attn_weights + bias_chunk

			max_score = jnp.max(attn_weights, axis=-1, keepdims=True)
			max_score = jnp.maximum(prev_max_score, max_score)
			max_score = jax.lax.stop_gradient(max_score)
			exp_weights = jnp.exp(attn_weights - max_score)
			exp_values = jnp.einsum(
				"bqhv,bvhd->bqhd", exp_weights, value_chunk, precision=precision
			)
			correction = jnp.exp(prev_max_score - max_score)
			numerator = numerator * correction + exp_values
			denominator = denominator * correction + exp_weights.sum(axis=-1, keepdims=True)
			return Carry(numerator, denominator, max_score), None

		def skip_upper_half(carry, args):
			key_chunk, value_chunk, key_chunk_idx = args
			skip_block = jnp.array(False)
			if causal:
				skip_block = query_chunk_idx < key_chunk_idx
			return jax.lax.cond(
				skip_block,
				lambda carry, args: (carry, None),
				scan_kv_block,
				carry,
				args,
			)

		init_carry = Carry(
			jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
			jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
			(-jnp.inf) * jnp.ones((batch, query_chunk_size, num_heads, 1), dtype=query.dtype),
		)
		(numerator, denominator, max_score), _ = lax.scan(
			skip_upper_half, init_carry, xs=(key, value, jnp.arange(0, num_kv))
		)
		outputs = (numerator / denominator).astype(dtype)
		return outputs

	_, res = lax.scan(
		lambda _, x: ((), scan_attention(x)), (), xs=(query, jnp.arange(0, num_q))
	)
	res = rearrange(res, "n b c h d -> b (n c) h d")
	return res


def _chunk_attention_bias(
	query_chunk_size: int,
	key_chunk_size: int,
	bias: chex.Array,
	deterministic: bool,
	attn_dropout: chex.Array,
	attention_drop_rate: float,
	causal: bool,
	dtype: chex.ArrayDType,  # type:ignore
	query_chunk_idx: int,
	key_chunk_idx: int,
):
	"""Helper function to compute the attention bias for a specific chunk."""
	query_offset = query_chunk_idx * query_chunk_size
	key_offset = key_chunk_idx * key_chunk_size
	chunk_bias = jnp.zeros((1, 1, 1, 1), dtype=dtype)
	if bias is not None:
		chunk_bias = lax.dynamic_slice(
			bias,
			start_indices=(0, 0, query_offset, key_offset),
			slice_sizes=(
				*bias.shape[:2],
				min(bias.shape[-2], query_chunk_size),
				min(bias.shape[-1], key_chunk_size),
			),
		)

	if causal:
		query_idx = lax.broadcasted_iota(
			dtype=jnp.int32, shape=(query_chunk_size, 1), dimension=0
		)
		key_idx = lax.broadcasted_iota(
			dtype=jnp.int32, shape=(1, key_chunk_size), dimension=1
		)
		offset = query_offset - key_offset
		query_idx += offset
		causal_mask_value = (query_idx < key_idx) * jnp.finfo(dtype).min
		chunk_bias += causal_mask_value.reshape(1, 1, *causal_mask_value.shape)

	if not deterministic and attention_drop_rate > 0.0:
		attn_dropout_slice = lax.dynamic_slice(
			attn_dropout,
			start_indices=(0, 0, query_offset, key_offset),
			slice_sizes=(
				*attn_dropout.shape[:2],
				min(attn_dropout.shape[-2], query_chunk_size),
				min(attn_dropout.shape[-1], key_chunk_size),
			),
		)
		chunk_bias += attn_dropout_slice * jnp.finfo(dtype).min
	return chunk_bias.astype(dtype)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base abstract class for all numerics."""

import abc
import flax.struct


class QNumerics(flax.struct.PyTreeNode, abc.ABC):
	"""Numerics for int8, int4, binary, etc."""

	# it in test. Remove and leave only get_dtype(

	@abc.abstractmethod
	def get_dtype(self):
		pass

	@abc.abstractmethod
	def fwd(self, x, context):
		"""Forward pass."""
		pass

	@abc.abstractmethod
	def abs_val_mapped_to(self):
		"""The value returned is the end of quantization range.

		It could be the biggest value that can be represented by numerical format
		exactly. E.g. in case of int8, 127 . Or it could be edge of the last bucket.
		Edge in case of int8, 127.5
		"""
		pass

	@abc.abstractmethod
	def vjp_fwd(self, x, context):
		pass

	@abc.abstractmethod
	def vjp_bwd(self, res, grad):
		pass



from fjformer.pallas_operations.tpu.flash_attention.flash_attention import (
	BlockSizes as BlockSizes,
)
from fjformer.pallas_operations.tpu.flash_attention.flash_attention import (
	flash_attention as flash_attention,
)



from fjformer.pallas_operations.tpu.splash_attention.splash_attention_kernel import (
	flash_attention_kernel as splash_flash_attention_kernel,
	SplashAttentionKernel as SplashAttentionKernel,
	BlockSizes as BlockSizes,
	attention_reference as attention_reference,
	SegmentIds as SegmentIds,
	make_splash_mha as make_splash_mha,
	make_splash_mqa as make_splash_mqa,
	make_masked_mha_reference as make_masked_mha_reference,
	make_splash_mha_single_device as make_splash_mha_single_device,
	make_splash_mqa_single_device as make_splash_mqa_single_device,
	make_masked_mqa_reference as make_masked_mqa_reference,
	QKVLayout as QKVLayout,
)

from fjformer.pallas_operations.tpu.splash_attention.splash_attention_mask import (
	Mask as Mask,
	LocalMask as LocalMask,
	make_local_attention_mask as make_local_attention_mask,
	make_causal_mask as make_causal_mask,
	make_random_mask as make_random_mask,
	FullMask as FullMask,
	NumpyMask as NumpyMask,
	CausalMask as CausalMask,
	MultiHeadMask as MultiHeadMask,
	LogicalOr as LogicalOr,
	LogicalAnd as LogicalAnd,
)

from fjformer.pallas_operations.tpu.splash_attention.splash_attention_mask_info import (
	MaskInfo as MaskInfo,
	process_mask as process_mask,
	process_mask_dkv as process_mask_dkv,
)

__all__ = (
	"MaskInfo",
	"process_mask",
	"process_mask_dkv",
	"Mask",
	"LocalMask",
	"make_local_attention_mask",
	"make_causal_mask",
	"make_random_mask",
	"FullMask",
	"NumpyMask",
	"CausalMask",
	"MultiHeadMask",
	"LogicalOr",
	"LogicalAnd",
	"splash_flash_attention_kernel",
	"SplashAttentionKernel",
	"BlockSizes",
	"attention_reference",
	"SegmentIds",
	"make_splash_mha",
	"make_splash_mqa",
	"make_masked_mha_reference",
	"make_splash_mha_single_device",
	"make_splash_mqa_single_device",
	"make_masked_mqa_reference",
	"QKVLayout",
)



# Why Do We have BITs here?

# Bits is a clone of https://github.com/google/aqt

# AQT library by google, but I needed to be able to change a lot of parts of the code and I couldn't do this on Google
# Repo for sure so I just made a copy of the part of the library that I wanted (AQT is Apache 2 Licenced Project)
# Ill more focus on doing the job for the llm checkpoint in 8 bit so there's no need to project name be AQT
# Accurate QTraining

from fjformer.bit_quantization.q_dot_general import (
	make_dot_general as make_dot_general,
	make_fake_quant as make_fake_quant,
	DotGeneralRes as DotGeneralRes,
)
from fjformer.bit_quantization.q_flax import (
	QuantMode as QuantMode,
	Freezer as Freezer,
	QDotGeneral as QDotGeneral,
	QEinsum as QEinsum,
	config_v4 as config_v4,
)

from fjformer.bit_quantization.config import fully_quantized

__all__ = (
	"make_dot_general",
	"make_fake_quant",
	"DotGeneralRes",
	"QuantMode",
	"Freezer",
	"QDotGeneral",
	"QEinsum",
	"config_v4",
	"fully_quantized",
)



from fjformer.pallas_operations.tpu.paged_attention.paged_attention import (
	MultiPageAsyncCopyDescriptor as MultiPageAsyncCopyDescriptor,
)
from fjformer.pallas_operations.tpu.paged_attention.paged_attention import (
	paged_attention as paged_attention,
)
